"""The python file for offset portfolio calculation."""
import logging
import math
import os
import json

from osgeo import ogr
import shapely
import shapely.speedups
import shapely.wkb
import shapely.prepared
import shapely.geometry
import shapely.geos

import adept_core

LOGGER = logging.getLogger('natcap.opal.offsets')

# offset schemes
OFFSET_SCHEME_ES = 0
OFFSET_SCHEME_BIODIV = 1
OFFSET_SCHEME_BIO_ES = 2

def build_shapely_polygon(ogr_feature, prep=False, fix=False):
    geometry = ogr_feature.GetGeometryRef()
    try:
        polygon = shapely.wkb.loads(geometry.ExportToWkb())
    except shapely.geos.ReadingError:
        LOGGER.debug('Attempting to close geometry rings')
        # If the geometry does not form a closed circle, try to connect the
        # rings with the OGR function.
        geometry.CloseRings()
        polygon = shapely.wkb.loads(geometry.ExportToWkb())
        LOGGER.debug('Geometry fixed')

    if fix:
        polygon = polygon.buffer(0)
    if prep:
        polygon = shapely.prepared.prep(polygon)
    return polygon

def distance_between_points(point_a, point_b):
    x1, y1 = point_a
    x2, y2 = point_b
    return math.hypot(math.fabs(x2 - x1), math.fabs(y2 - y1))

def poly_intersects_vector(polygon, vector_uri):
    """Simple function to check if a polygon intersects any polygon in the
    input vector.  Returns a boolean True or False.

        polygon - a shapely polygon
        vector_uri - a URI to an OGR vector.

    Returns True or False."""

    vector = ogr.Open(vector_uri)
    layer = vector.GetLayer()

    for feature in layer:
        shapely_feature = build_shapely_polygon(feature, prep=True)
        if shapely_feature.intersects(polygon):
            return True
    return False


def locate_parcels(possible_parcels, selection_area, impact_sites):
    """This function will iterate through all polygons in the possible_parcels
    vector that are within the selection area.

        possible_parcels - a URI to an OGR vector of polygons.
        selection_area - a URI to an OGR vector of the area to consider.
        impact_sites - a URI to an OGR vector of impact sites.

    #TODO: What should this function return?
    Returns something ... not yet sure what."""

    # TODO: need a way to prioritize polygons.  Factors to consider:
    #  - distance of offset parcel centroid from the impact parcel. (prefer nearer)
    #  - whether the patch contributes to biodiversity (prefer if yes)
    #    - if biodiversity parcel < 5ha, biodiversity from this parcel is not
    #      counted towards the total biodiversity offset.
    #  - parcel size.

    # TODO: start out with just selecting all the parcels completely contained
    # within the selection area.

    if shapely.speedups.available:
        LOGGER.debug('Enabling shapely speedups')
        shapely.speedups.enable()

    LOGGER.debug('Extracting selection area polygon')
    sa_vector = ogr.Open(selection_area)
    sa_layer = sa_vector.GetLayer()
    sa_feature = sa_layer.GetFeature(0)
    sa_geometry = sa_feature.GetGeometryRef()
    sa_polygon = shapely.wkb.loads(sa_geometry.ExportToWkb())
    #selection_area = shapely.geometry.polygon.Polygon(sa_polygon)

    # Open the impact sites vector
    LOGGER.debug('Opening the impact sites vector')
    impact_vector = ogr.Open(impact_sites)
    impact_layer = impact_vector.GetLayer()

    # create a data structure that allows me to prioritize possible polygons.
    parcels = []

    LOGGER.debug('Opening the parcels vector')
    ecosystems_vector = ogr.Open(possible_parcels)
    ecosystems_layer = ecosystems_vector.GetLayer()
    for possible_parcel in ecosystems_layer:
        # see if the parcel intersect with the selection_area
        polygon = build_shapely_polygon(possible_parcel)
        prepared_polygon = shapely.prepared.prep(polygon)
        if prepared_polygon.intersects(sa_polygon):
            LOGGER.debug('Possible offset area intersects.')

            distances = {}
            for impact_id, ogr_impact_site in enumerate(impact_layer):
                impact_site = build_shapely_polygon(ogr_impact_site)
                key = 'impact_%s' % impact_id

                # currently accessing the minimum distance between polygons.  If we
                # want the distance between centroids, get the point's coordinates
                # like so:
                # impact_site.centroid.coords[0] and pass it to the
                # distance_between_points function, above.
                distance = impact_site.distance(polygon)
                distances[key] = distance
            impact_layer.ResetReading()

            offset_parcel_data = {
                'polygon': polygon,
                'area': polygon.area,
                'distances':  distances,
            }
            parcels.append(offset_parcel_data)

    #print parcels

def _select_offsets(offset_parcels_uri, impact_sites_uri, biodiversity_impacts, output_vector,
        output_json, comparison_vectors={}, services=['carbon', 'nutrient', 'sediment'],
        offset_factors={'carbon': 1.0, 'nutrient': 1.0,'sediment': 1.0, 'biodiversity': 1.0},
        offset_scheme=OFFSET_SCHEME_BIODIV):
    """Determine which parcels to return, where parcels returned meet
    appropriate criteria defined in the inputs.

    offset_parcels_uri - a URI to a vector containing polygons.  Parcels
        provided here are the parcels available for use as offsets.  Parcels
        must have the following attributes:
            ecosystem - a string with the ecosystem name
            LCI - the Landscape Context Index for the parcel (float)
            carbon - (float) only required if carbon in services input.
            nutrient - (float) only required if nutrient in services input.
            sediment - (float) only required if sediment in services input.
    impact_sites_uri - a URI to a vector containing polygons.  Parcels provided
        here are those that have been analyzed as impacts to the landscape.
        Parcels must have these columns:
            carbon - (float) only required if carbon in services input.
            nutrient - (float) only required if nutrient in services input.
            sediment - (float) only required if sediment in services input.
            Threat - (float) required if an ecosystem has threat data.
            Richness - (float) required if an ecosystem has richness data.
    biodiversity_impacts - Dict mapping string ecosystem names to dicts with
        the following key-value pairs:
            min_impacted_parcel_area - (number) The area (in Ha) of the smallest
                impacted natural parcel of this ecosystem type.
            min_lci - (number between 0, 1) The lowest LCI of all impacted
                parcels of this ecosystem type.
            max_threat - (number) The max threat value of all impacted natural
                parcels of this ecosystem type.  This will be None if there is
                no threat data.
            min_richness - (number) The min richness value of all impacted
                natural parcels of this ecosystem type.  This will be None if
                there is no richness data.
            mitigation_area - (number) The area (in ha) required for mitigating
                the impacts to this ecosystem type.
    output_vector - a URI to where an output vector should be written.  This
        vector will contain a subset of the parcels provided in the vector at
        offset_parcels_uri, where subset parcels are those that have been
        'selected' by this function.
    output_json - a URI to where an output JSON object should be written.  This
        JSON object will contain detailed information about the parcels that
        have been selected by this function.
    comparison_vectors - a dict mapping string keys to URIs representing OGR
        vectors on disk.  These vectors are used to determine whether offset
        polygons intersect these vectors.  The resulting binary information is
        recorded using the string key in the output vector and JSON object.
    services - a list of string ecosystem service names.  Any strings in this
        list must also exist in the impact_sites_uri vector as columns, where
        the column data is numeric (usually a float).
    offset_factor - a number (float or int) indicating the proportion of the
        the total required offsets that the set of offset parcels selected
        should offset.  Default: 1.0.
    offset_scheme - a number, one of OFFSET_SCHEME_ES, OFFSET_SCHEME_BIODIV, or
        OFFSET_SCHEME_BIO_ES.  Indicates which offset scheme should be used.

    Returns a tuple.  The first tuple element is a list of all integer parcel
    indices selected.  The second tuple element is a set of recommended parcels
    based on known impacts/offsets.
    """

#        offset_factor=1.0 - an int or float.  Offset parcels will be selected
#            for each ecosystem type until the ecosystem type's
#            mitigation_area*offset_factor is met or we run out of parcels to
#            select.
    # comparison_vectors should have string keys mapping to URIs.
    LOGGER.debug('Opening impacts sites vector %s', impact_sites_uri)
    impact_vector = ogr.Open(impact_sites_uri)
    impact_layer = impact_vector.GetLayer()
    impact_schema = map(lambda d: d.GetName(), impact_layer.schema)

    # check that all fields passed in via services parameter are found in the
    # impact sites vector.  It's an error if not.
    for service_name in services:
        if service_name not in impact_schema:
            raise RuntimeError('Service %s not defined in impact schema' %
                service_name)

    # Get information about which offset parcels we can actually use for
    # mitigation.
    LOGGER.debug('Getting possible offset parcels')
    if len(biodiversity_impacts) == 0 or\
        offset_scheme in [OFFSET_SCHEME_ES, OFFSET_SCHEME_BIO_ES]:
        include_all_ecosystems = True
    else:
        include_all_ecosystems = False
    bio_offsets = locate_biodiversity_offsets(offset_parcels_uri,
        biodiversity_impacts, include_all_ecosystems)

    # loop through the impact sites and get a list of centroid coordinates
    impact_polygons = [build_shapely_polygon(site).centroid for site in impact_layer]

#    for offset_index in xrange(num_offset_parcels):
#        offset_feature = offsets_layer.GetFeature(offset_index)
#        offset_polygon = build_shapely_polygon(offset_feature)
#        ecosystem_impacted = offset_feature.GetField('ecosystem')
#        parcel_area = offset_feature.GetGeometryRef().Area()
#        parcel_lci = offset_feature.GetField('LCI')
#

    LOGGER.debug('Calculating distance from offset to impact sites')
    offsets_vector = ogr.Open(offset_parcels_uri)
    offsets_layer = offsets_vector.GetLayer()

    for offset_index, offset_data in bio_offsets.iteritems():
        offset_parcel = offsets_layer.GetFeature(offset_index)
        offset_polygon = build_shapely_polygon(offset_parcel)

        # figure out if our offset polygon intersects any of the comparison
        # vectors.
        for comp_name, comp_vector in comparison_vectors.iteritems():
            intersects = poly_intersects_vector(offset_polygon, comp_vector)
            bio_offsets[offset_index][comp_name] = intersects

        offset_polygon = offset_polygon.centroid

        min_distance = min(map(lambda x: offset_polygon.distance(x),
            impact_polygons))
        bio_offsets[offset_index]['distance'] = min_distance

    # for now, just sort based on distance from impact parcel.
    sorted_offset_parcels = sorted(bio_offsets.keys(),
        key=lambda k: bio_offsets[k]['distance'])
    LOGGER.debug('Sorted offset parcels: %s', sorted_offset_parcels)

    # select enough parcels to offset the impact requirements and still meet
    # the minimum number of parcels required.

    # impact_offsets tracks the amt of area offset per ecosystem type.
    # initialize all the ecosystems tracked to 0.
    impact_offsets = dict((e_n, 0.0) for e_n in list(set(map(
        lambda x: bio_offsets[x]['ecosystem'], bio_offsets.keys()))))
    LOGGER.debug('impact_offsets: %s', impact_offsets)

    min_parcels_to_select = 20
    min_offset_factor = offset_factors['biodiversity']
    parcels_selected = {}
    new_field_data = {}
    new_fields = ['distance'] + comparison_vectors.keys()
    recommendations = {}
    biodiversity_requirements = {}
    for ecosystem_name, offset_amt in impact_offsets.iteritems():
        # Richness and threat data are stored as None or Floats.
        # In python, None is less than -inf.
        ecosys_in_bio_impacts = ecosystem_name in biodiversity_impacts
        if len(biodiversity_impacts) > 0 and ecosys_in_bio_impacts:
            max_threat = biodiversity_impacts[ecosystem_name]['max_threat']
            min_richness = biodiversity_impacts[ecosystem_name]['min_richness']

            # required mitigation area is in ha, yet parcel areas are in m^2
            # multiply ha by 10,000 to get m^2
            required_mitigation_area = biodiversity_impacts[ecosystem_name]['mitigation_area']
            min_mitigation = (required_mitigation_area * 10000) * min_offset_factor
        else:
            # when there are no biodiversity impacts, we need to assume some
            # default values so that we only account for ecosystem services.
            max_threat = None
            min_richness = None
            required_mitigation_area = 0.0
            min_mitigation = 0.0

        LOGGER.debug('Locating offsets for ecosystem "%s"', ecosystem_name)
        parcels_in_ecosystem = []
        for possible_offset in sorted_offset_parcels:
            parcel_feature = offsets_layer.GetFeature(possible_offset)

            if parcel_feature.GetField('ecosystem') == ecosystem_name:
                # if we have Threat data, we need to enforce that all offset
                # parcels meet the minimum threat score requirement.
                if max_threat is not None:
                    parcel_threat = parcel_feature.GetField('Threat')
                    if parcel_threat > max_threat:
                        # if threat requirement is not met, skip this parcel
                        continue

                if min_richness is not None:
                    parcel_richness = parcel_feature.GetField('Richness')
                    if parcel_richness < min_richness:
                        # if the richness requirement is not met, skip this
                        # parcel
                        continue

                parcels_in_ecosystem.append(possible_offset)

                parcel_area = parcel_feature.GetGeometryRef().Area()
                parcels_selected[possible_offset] = parcel_area
                impact_offsets[ecosystem_name] += parcel_area

                parcel_data = {}
                for fieldname in new_fields:
                    new_field_value = bio_offsets[possible_offset][fieldname]
                    parcel_data[fieldname] = new_field_value

                new_field_data[possible_offset] = parcel_data
                new_field_data[possible_offset]['area'] = parcel_area
                new_field_data[possible_offset]['ecosystem_type'] = ecosystem_name
                new_field_data[possible_offset]['parcel_id'] = possible_offset

                for es_field in services:
                    es_value = parcel_feature.GetField(es_field)
                    new_field_data[possible_offset][es_field] = es_value

                # Uncomment these lines if you want to only show the minimum
                # number of parcels needed to meet the rules above.
#                if impact_offsets[ecosystem_name] >= min_mitigation and \
#                    len(parcels_selected) >= min_parcels_to_select:
#                    LOGGER.debug('Parcel requirements have been satisfied.')
#                    break

        # I only want those offsets that match this ecosystem name
        available_offset_parcels = {}
        for parcel in parcels_in_ecosystem:
            available_offset_parcels[parcel] = new_field_data[parcel]

#        TODO: add requirements to the requirements list depending on whether
#        we want to include biodiversity and/or ES restrictions when selecting
#        a set of parcels.
#        recommended_parcels = select_set(available_offset_parcels,
#            [('area', min_mitigation)])
#        recommendations[ecosystem_name] = recommended_parcels
#        LOGGER.debug('Recommended parcels for %s: %s', ecosystem_name,
#            recommended_parcels)
#
#    LOGGER.debug('Found %s offset parcels to meet requirements',
#            len(parcels_selected))

    offset_parcels_schema = map(lambda d: d.GetName(), offsets_layer.schema)
    output_fields = ['ecosystem'] + services
    new_fields += ['parcel_id', 'area']
    field_types = {
        'parcel_id': ogr.OFTInteger,
        'Subzone': ogr.OFTInteger,
        'AOI': ogr.OFTInteger,
        'City': ogr.OFTInteger,
        'distance': ogr.OFTReal,
        'area': ogr.OFTReal,
    }
    for key in ['LCI', 'Threat', 'Richness', 'custom']:
        if key in offset_parcels_schema:
            output_fields.append(key)

    index_map = adept_core.write_vector(offset_parcels_uri, parcels_selected.keys(),
        output_vector, output_fields, new_fields, new_field_data, field_types)

    updated_recommendations = {}
    for ecosystem_name, suggested_fid in recommendations.iteritems():
        new_fids = []
        for fid in suggested_fid:
            new_fids.append(index_map[fid])

        updated_recommendations[ecosystem_name] = new_fids

    json.dump({'field_data': new_field_data,
        'parcel_sets': updated_recommendations},
        open(output_json, 'w'), sort_keys=True, indent=4)

    return (parcels_selected.keys(), updated_recommendations)

def locate_biodiversity_offsets(offset_parcels_uri, biodiversity_impacts,
        include_all_ecosystems=False):
    """Given a vector of possible offset parcels and a set of known impacts,
    determine which offset parcels can be considered.  An offset parcel will be
    considered if its ecosystem type is in the set of known, impacted ecosystem
    types.

        offset_parcels_uri - a URI to an OGR vector containing offset parcels.
            This vector must include the following attributes:
                ecosystem - A string ecosystem name
                LCI - the Landscape Context Index for this parcel (float, 0-1)
        biodiversity_impacts - a dict with keys being string ecosystem names
            that may be found in the ecosystem column of the offset parcels
            vector.  Values are dicts with at least these attributes:
                'min_impacted_parcel_area' - a number
                'min_lci' - a number
        include_all_ecosystems=False - a boolean.  If True, the resulting
            dictionary will include all ecosystems encountered.  Biodiversity
            offsets will still be reported as normal, but the output dictionary
            will include ecosystems that are not valid biodiversity offsets.
            These parcels that are not valid biodiversity offsets will have
            values of 0 or None.

    Under normal conditions, this function will only look at ecosystem types
    that are found in the biosiversity_impacts dict.  In the case where
    biodiversity_impacts contains no entries, ALL ecosystem types will be
    considered.

    Returns a dict mapping integer parcel indices to dict values with the
    following attributes:
        ecosystem - the string ecosystem name of the offset parcel
        area - the area of the offset parcel
        lci - the Landscape Context Index of the offset parcel"""

    LOGGER.debug('Opening parcels vector %s', offset_parcels_uri)
    offsets_vector = ogr.Open(offset_parcels_uri)
    offsets_layer = offsets_vector.GetLayer()
    offsets_srs = offsets_layer.GetSpatialRef()
    offsets_schema = set(map(lambda d: d.GetName(), offsets_layer.schema))
    LOGGER.debug(offsets_schema)

    service_funcs = {}
    known_services = {
        'carbon': lambda f: f.GetField('carbon'),
        'sediment': lambda f: f.GetField('sediment'),
        'nutrient': lambda f: f.GetField('nutrient'),
        'custom': lambda f: f.GetField('custom'),
    }
    for service_name, service_field_func in known_services.iteritems():
        if service_name in offsets_schema:
            service_funcs[service_name] = service_field_func

    LOGGER.debug(service_funcs.keys())

    valid_offset_parcels = {}

    # if we aren't doing any biodiversity restrictions, allow parcels from ANY
    # ecosystem to be included.
    use_any_ecosystem = len(biodiversity_impacts) == 0 or include_all_ecosystems

    num_offset_parcels = offsets_layer.GetFeatureCount()
    LOGGER.debug('Examining %s possible offset parcels', num_offset_parcels)
    for offset_index in xrange(num_offset_parcels):
        offset_feature = offsets_layer.GetFeature(offset_index)
        ecosystem_impacted = offset_feature.GetField('ecosystem')

        # Convert parcel area to Ha.
        # eco_impact_data['min_impacted_parcel_area'] is also in Ha.
        parcel_area = offset_feature.GetGeometryRef().Area() / 10000.0

        try:
            parcel_lci = offset_feature.GetField('LCI')
        except ValueError:
            parcel_lci = None

        print "%s, %s" % (parcel_area, parcel_lci)

        if ecosystem_impacted in biodiversity_impacts:
            eco_impact_data = biodiversity_impacts[ecosystem_impacted]
            min_area = eco_impact_data['min_impacted_parcel_area']
            min_lci = eco_impact_data['min_lci']
            if parcel_area >= min_area and parcel_lci >= min_lci:
                offset_parcel_data = {
                    'ecosystem': ecosystem_impacted,
                    'area': parcel_area,
                    'lci': parcel_lci,
                }
                valid_offset_parcels[offset_index] = offset_parcel_data
        elif use_any_ecosystem is True:
            valid_offset_parcels[offset_index] = {
                'ecosystem': ecosystem_impacted,
                'area': parcel_area,
                'lci': parcel_lci,
            }

    LOGGER.debug('Found %s possible offset parcels out of %s provided',
        len(valid_offset_parcels), num_offset_parcels)

    LOGGER.debug(json.dumps(valid_offset_parcels, indent=4, sort_keys=True))
    return valid_offset_parcels

def select_offsets(offset_parcels_uri, impact_sites_uri, output_dir):
    """select offset sites for each impact site.  All sites in the offset
    parcels are available for consideration.

    This function will create one new OGR vector per impact site.  Each of
    these offset vectors will contain the selected geometries that offset the
    impact.  An offset parcel may not offset more than one impact site.

    offset vectors will be saved to the output_dir.

    Returns a dictionary with this structure:
        {
            impact site ID: {
                service_name_1: service mitigated
                service_name_2: service mitigated
            }
            ...
        }"""

    if shapely.speedups.available:
        LOGGER.debug('Enabling shapely speedups')
        shapely.speedups.enable()

    LOGGER.debug('Opening parcels vector %s', offset_parcels_uri)
    parcels_vector = ogr.Open(offset_parcels_uri)
    parcels_layer = parcels_vector.GetLayer()
    parcels_srs = parcels_layer.GetSpatialRef()

    LOGGER.debug('Opening impact vector %s', impact_sites_uri)
    impact_vector = ogr.Open(impact_sites_uri)
    impact_layer = impact_vector.GetLayer()

    if not os.path.exists(output_dir):
        LOGGER.debug('Creating output folder %s', output_dir)
        os.makedirs(output_dir)

    # build up a data structure to represent the available offset polygons
    # any parcels represented in available_parcels are available for selection.
    all_parcels = {}
    for parcel in parcels_layer:
        parcel_fid = parcel.GetFID()  # feature identifier
        parcel_data = {
            'polygon': build_shapely_polygon(parcel),
            'sediment': parcel.GetField('sediment'),
            'nutrient': parcel.GetField('nutrient'),
            'carbon': parcel.GetField('carbon'),
            'offsets': None,  # if this is an int, that's the impact site fid
        }
        all_parcels[parcel_fid] = parcel_data

    offsets_record = {}
    for impact_site in impact_layer:
        # get the impact site's geometry and convert to a shapely polygon.
        impact_polygon = build_shapely_polygon(impact_site)
        impact_site_fid = impact_site.GetFID()
        impact_quantity = impact_site.GetField('sediment')
        quantity_offset = 0.0
        out_vector_uri = None  # if None, no file written with portfolio

        if impact_quantity == 0:
            LOGGER.info('Impact site %s has nothing to offset.',
                    impact_site_fid)
        else:
            available_parcels = []
            for parcel_fid, parcel_data in all_parcels.iteritems():
                if parcel_data['offsets'] is None:  # parcel doesn't offset something
                    available_parcels.append({
                        'fid': parcel_fid,
                        'sediment': parcel_data['sediment'],
                        'nutrient': parcel_data['nutrient'],
                        'carbon': parcel_data['carbon'],
                        'distance': impact_polygon.distance(parcel_data['polygon']),
                    })

            # sort the available offset parcels.  This could easily be replaced by
            # a more sophisticated way of prioritizing offset parcels.  For now,
            # I'm relying solely on the minimum distance between impact and offset
            # sites.
            sorted_offset_parcels = sorted(available_parcels, key=lambda p:
                    p['distance'])
            LOGGER.debug('Evaluating %s parcels', len(sorted_offset_parcels))
            parcels_selected = []  # bookkeeping list

            # select offset parcels for this impact site.
            for offset_parcel in sorted_offset_parcels:
                if quantity_offset >= impact_quantity:
                    LOGGER.debug('Impact site %s offsets met.', impact_site_fid)
                    break  # we've offset everything we want/need

                # we can only select the parcel if it doesn't intersect with the
                # impact parcel.
                # get the parcel's geometry, see if it intersects impact_polygon.
                #offset_polygon = build_shapely_polygon(
                #    parcels_layer.GetFeature(offset_parcel['fid']))
                if offset_parcel['distance'] > 0:  # does not intersect
                #if not offset_polygon.intersects(impact_polygon):
                    # mark the parcel as selected
                    offset_parcel_fid = offset_parcel['fid']
                    all_parcels[offset_parcel_fid]['offsets'] = impact_site_fid
                    parcels_selected.append(offset_parcel_fid)

                    # increase the per-impact service that's been offset
                    quantity_offset += offset_parcel['sediment']

            LOGGER.debug('Finished evaluating impact site %s', impact_site_fid)
            if quantity_offset < impact_quantity:
                if quantity_offset == 0.:
                    percent_offset = 0
                else:
                    percent_offset = impact_quantity/quantity_offset
                LOGGER.warn('Impact site %s only offset %s percent of its impact',
                    impact_site_fid, percent_offset)

            # once we've selected all of our parcels in the previous loop, write
            # these new polygons to a new vector in the user's defined output
            # folder.  If no offser parcels have been selected for this impact
            # site, don't create an output vector for it.
            if len(parcels_selected) == 0:
                LOGGER.debug('Impact site %s has no offsets', impact_site_fid)
            else:
                new_vector_name = "site_%s_offsets.shp" % impact_site_fid
                layer_name = str(new_vector_name[:8])  # shapefile requires ASCII
                out_vector_uri = os.path.join(output_dir, new_vector_name)

                out_driver = ogr.GetDriverByName('ESRI Shapefile')
                out_vector = out_driver.CreateDataSource(out_vector_uri)

                out_layer = out_vector.CreateLayer(layer_name, srs=parcels_srs)
                # create a new offset field for each ES.
                es_fields = ['sediment', 'nutrient', 'carbon']
                for field_name in es_fields:
                    field_defn = ogr.FieldDefn(field_name, ogr.OFTReal)
                    out_layer.CreateField(field_defn)
                out_layer_defn = out_layer.GetLayerDefn()

                for parcel_fid in parcels_selected:
                    parcel_feature = parcels_layer.GetFeature(parcel_fid)
                    parcel_geom = parcel_feature.GetGeometryRef()
                    parcel_defn = parcel_feature.GetDefnRef()

                    new_feature = ogr.Feature(parcel_defn)
                    new_feature.SetGeometry(parcel_geom)

                    # fetch the ES data from the parcel and copy it over.
                    for field_name in es_fields:
                        parcel_data = parcel_feature.GetField(field_name)
                        field_index = out_layer_defn.GetFieldIndex(field_name)
                        new_feature.SetField(field_index, parcel_data)

                    out_layer.CreateFeature(new_feature)

                out_layer.SyncToDisk()
                out_layer = None
                out_vector = None
                out_driver = None

        # For the sake of consistency, check if the impact quantity has not
        # been set.  If this is the case, set the impact quantity to 0.0
        if impact_quantity is None:
            impact_quantity = 0.0

        offsets_record[impact_site_fid] = {
                'offset_portfolio': out_vector_uri,
                'services': {
                    'sediment': {
                        'impact': impact_quantity,
                        'offset_sum': quantity_offset,
                    },
                }
            }
    offsets_json_uri = os.path.join(output_dir, 'offsets.json')
    LOGGER.debug('Saving offset summary to %s', offsets_json_uri)
    json.dump(offsets_record, open(offsets_json_uri, 'w'), sort_keys=True, indent=4)
    return offsets_record

def select_set(parcels_dict, requirements):
    """Select a set of parcels that together meet a set of minimum
    requirements, outlined in the `requirements` input.

        parcels_dict - a dictionary describing the available parcels and the
            various attributes we're considering in parcel selection.
            example:
                parcels_dict = {
                    1: {'area': 400, 'carbon': 3500, 'sediment': 290},
                    2: {'area': 123, 'carbon': 382, 'sediment': 1348},
                    3: {'area': 8392, 'carbon': 1910, 'sediment': 18234},
                    4: {'area': 149, 'carbon': 192, 'sediment': 1019},
                }
            The key is the parcel identifier.  It can be of any type.
            The keys of the parcel dictionary ('area', 'carbon', 'sediment')
            can be any keys, so long as they match the service names in the
            requirements list.  The values of the parcel dictionary MUST be
            numeric.  They represent the service value for this parcel.
        requirements - a list of tuples.
            example:
                requirements = [
                    ('area', 350),
                    ('sediment', 2000),
                    ('carbon', 100),
                ]
            The first element of the tuple is the service name.  It must match
            the one of the service names found in the parcels_dict.
            The second element is the minimum required amount of this service
            for parcel selection.

        In the examples given, Parcel 3 meets all three requirements, so it is
        the only parcel returned from this function.

        Returns a sorted list of parcel identifiers meeting the above
        requirements."""

    # use sets because a parcel ID can only appear once and order doesn't
    # matter, and constant-time accesses are useful.
    available_parcels = set(parcels_dict.keys())
    recommended_parcels = set()

    for rule_key, rule_min in requirements:
        # sort parcels from max to min.  We want a minimum number of parcels
        # that meets this requirement.
        # available_parcels contains only those parcels that have not been
        # selected.
        sorted_parcels = sorted(available_parcels,
            key=lambda pid: parcels_dict[pid][rule_key], reverse=True)

        rule_sum = 0
        if len(recommended_parcels) > 0:
            # If we already have selected parcels, we need to first check if
            # they already satisfy this requirement.
            rule_sum = sum([parcels_dict[x][rule_key] for x in recommended_parcels])
            if rule_sum >= rule_min:
                LOGGER.debug(('Rule %s already met with existing parcel set'
                    ' (min=%s, sum=%s)'), rule_key, rule_min, rule_sum)
                continue

        for parcel_id in sorted_parcels:
            if rule_sum >= rule_min:
                LOGGER.debug('Met rule %s >= %s with %s', rule_key, rule_min,
                    rule_sum)
                break
            rule_sum += parcels_dict[parcel_id][rule_key]
            recommended_parcels.add(parcel_id)
            available_parcels.remove(parcel_id)
        LOGGER.debug('Parcel set so far: %s', recommended_parcels)

    LOGGER.debug('Found %s parcels to recommend', len(recommended_parcels))
    return sorted(recommended_parcels)


# quick function to convert from m to ha
Ha = lambda m: m / 10000.0

def select_set_multifactor(parcels, biodiversity_req=None, es_hydro_req=None,
        es_global_req=None, proportion_offset=1.0):
    """Multifactor parcel selection.  Selects offset parcels based on
    biodiversity, hydrological ES and global ES requirements.

    parcels - a dict mapping integer offset parcel IDs to dictionaries with
        offset parcel information:
            'area' - a numberic value, the parcel's area.  Required if
                biodiversity_req is provided.
            'ecosystem' - a string ecosystem name for this offset parcel.
                Required if biodiversity_req is provided.
    biodversity_req - a dictionary mapping string ecosystem names to a
        dictionary with the following entrie(s):
            'mitigation_area' - the numeric area required to offset impacts to
                this ecosystem type.
    proportion_offset=1.0 - a number indicating the proportion of
        impacts/requirements to offset.
"""
    if biodiversity_req is None and es_hydro_req is None and es_global_req is None:
        raise Exception('No requirements specified')
    proportion_offset = float(proportion_offset)
    LOGGER.debug('Offsetting with proportion = %s', proportion_offset)

    selected_parcels = set([])
    if biodiversity_req is not None:
        # determine which parcels are in which ecosystem.
        ecosystems = {}
        for parcel_key, parcel_data in parcels.iteritems():
            ecosystem_id = parcel_data['ecosystem']
            try:
                ecosystems[ecosystem_id].append(parcel_key)
            except:
                ecosystems[ecosystem_id] = [parcel_key]

        for impacted_ecosystem, ecosys_data in biodiversity_req.iteritems():
            required_area = ecosys_data['mitigation_area']
            ecosys_selected_area = 0
            ecosystem_selected_parcels = []
            try:
                possible_parcels = ecosystems[impacted_ecosystem]
            except KeyError:
                LOGGER.warn(('Impacted ecosystem "%s" does not have any '
                    'available offset parcels') % impacted_ecosystem)
                continue

            # sort so that largest parcel areas are first.
            sorted_parcels = sorted(possible_parcels,
                key=lambda p: parcels[p]['area'], reverse=True)

            for parcel_key in sorted_parcels:
                if ecosys_selected_area >= (required_area * proportion_offset):
                    LOGGER.debug('Ecosystem %s area req. (%s*%s) met with (%s)',
                        impacted_ecosystem, proportion_offset, required_area,
                        ecosys_selected_area)
                    break

                parcel_data = parcels[parcel_key]

                # select the parcel.
                ecosystem_selected_parcels.append(parcel_key)
                ecosys_selected_area += parcel_data['area']

            # add the newly selected parcels to the running set
            selected_parcels.update(ecosystem_selected_parcels)
        LOGGER.debug('Parcels selected after biodiversity: %s',
            selected_parcels)

    if es_hydro_req is not None:
        hydro_selected_parcels = set([])
        for serviceshed_id, serviceshed_data in es_hydro_req.iteritems():
            try:
                sshed_parcels = serviceshed_data['parcels']
            except KeyError:
                # If there are no parcels in this serviceshed, skip
                # the serviceshed.  Related to adept_core.py, line 667.
                continue

            # Figure out which hydrological services have impacts and if we
            # have selected parcels, we want to know the service offset values
            # of those parcels.
            services = {}
            for service_key in ['sediment', 'nutrient', 'custom']:
                try:
                     sshed_service_data = {
                        'required_amt': serviceshed_data[service_key],
                        'offset': 0
                    }
                except KeyError:
                    # when the service is not in serviceshed_data, we're not
                    # trying to meet that service's requrements.
                    continue

                for sshed_parcel in sshed_parcels:
                    if sshed_parcel in selected_parcels:
                        service_offset = parcels[sshed_parcel][service_key]
                        sshed_service_data['offset'] += service_offset
                LOGGER.debug('sshed %s %s: %s', serviceshed_id, service_key,
                    sshed_service_data['offset'])

                # keep track of the service data for this serviceshed.
                services[service_key] = sshed_service_data

            # select available parcels until requirements have been met.
            remaining_parcels = set(serviceshed_data['parcels']).difference(
                selected_parcels)
            LOGGER.debug('Remaining parcels: %s', remaining_parcels)
            if len(remaining_parcels) == 0:
                continue

            # iterate through the provided services in alphabetical order.
            for service, service_data in sorted(services.iteritems(),
                    key=lambda x: x[0]):
                # get a list of available parcels, sorted in increasing order
                offset_parcels = sorted(remaining_parcels, key=lambda x:
                    parcels[x][service], reverse=True)
                LOGGER.debug('Offset parcels: %s', offset_parcels)
                service_offset = service_data['offset']
                required_amt = service_data['required_amt'] * proportion_offset

                for offset_parcel in offset_parcels:
                    if service_offset > required_amt:
                        break

                    # if we've already picked the parcel, skip it.
                    if offset_parcel in hydro_selected_parcels:
                        continue

                    parcel_data = parcels[offset_parcel]
                    service_offset += parcel_data[service]

                    # account for all of the services that parcel provides
                    for _service in services.keys():
                        services[_service]['offset'] += parcel_data[_service]
                    remaining_parcels.remove(offset_parcel)
                    hydro_selected_parcels.add(offset_parcel)
                    selected_parcels.add(offset_parcel)
                    LOGGER.debug('selected parcels: %s', selected_parcels)

    if es_global_req is not None:
        # Determine the service offsets for any parcels we've already selected.
        service_totals = {}
        for service in es_global_req.keys():
            if len(selected_parcels) > 0:
                service_init = 0
            else:
                service_init = sum(map(lambda pid:
                    parcels[pid][service], selected_parcels))
            service_totals[service] = service_init

        available_parcels = set(parcels.keys()).difference(selected_parcels)
        global_es_selected_parcels = set([])
        for service in sorted(es_global_req.keys()):
            # sort available parcels in decreasing order of service offset
            service_sort = lambda pid: parcels[pid][service]
            for parcel_id in sorted(available_parcels, key=service_sort,
                reverse=True):
                amt_required = es_global_req[service] * proportion_offset
                if service_totals[service] >= amt_required:
                    break

                available_parcels.remove(parcel_id)
                global_es_selected_parcels.add(parcel_id)
                for _service in es_global_req.keys():
                    service_totals[_service] += parcels[parcel_id][_service]

        LOGGER.debug('Global selected parcels: %s', global_es_selected_parcels)
        selected_parcels.update(global_es_selected_parcels)

    return sorted(selected_parcels)

def translate_parcel_data(per_offset_data):
    """Translate the per-offset parcel data dictionary returned from
    natcap.opal.analysis.percent_overlap() into the data structures required for
    multifactor parcel selection to take place."""
    output_parcels = {}
    output_ssheds = {}
    known_keys = {
        'Carbon': 'carbon',
        'Nitrogen': 'nutrient',
        'Sediment': 'sediment',
        'Custom': 'custom',
        'Area': 'area',
        'Ecosystem': 'ecosystem',
        'municipalities': 'overlap',
    }
    for parcel_id, parcel_data in per_offset_data.iteritems():
        output_parcel_data = {}
        for old_key, new_key in known_keys.iteritems():
            try:
                value = parcel_data[old_key]
                if new_key == 'area':
                    value = Ha(value)  # convert from m^2 to Ha
                output_parcel_data[new_key] = value
            except KeyError:
                # When old_key isn't in parcel_data
                pass
        output_parcels[parcel_id] = output_parcel_data

    return output_parcels

def translate_es_impacts(per_impact_data, custom_type='global'):
    """Translate the per-impact data dictionary returned from
    natcap.opal.analysis.percent_overlap() into the hydrological ES dictionary.

    Returns one dictionary for hydrological impacts, and another one for global
    impacts."""

    # structure:
    #  'sshed_name': {
    #    'service': required_service_value, ...}
    #
    output_hydro_impacts = {}
    output_global_impacts = {}
    known_keys = {
        'carbon': 'Carbon',
        'nutrient': 'Nitrogen',
        'sediment': 'Sediment',
        'custom': 'Custom',
    }
    services = set(['carbon', 'nutrient', 'sediment', 'custom'])
    for parcel_id, parcel_data in per_impact_data.iteritems():
        for sshed_name, sshed_overlap in parcel_data['municipalities'].iteritems():
            # initialize values if necessary
            if sshed_name not in output_hydro_impacts:
                sshed_services = {}
                for service in services:
                    if known_keys[service] not in parcel_data:
                        continue
                    sshed_services[service] = 0
                output_hydro_impacts[sshed_name] = sshed_services

            for service in services:
                try:
                    # TODO: if sshed_overlap < 1, allocate remaining impact to
                    # global requirements.
                    service_impact = parcel_data[known_keys[service]] * sshed_overlap
                except (KeyError, TypeError):
                    # KeyError when service is not in parcel_data
                    # TypeError when the parcel_data[service] value is None
                    # in all cases, skip it.
                    continue

                if sshed_name not in output_hydro_impacts:
                    output_hydro_impacts[sshed_name] = {}
                output_hydro_impacts[sshed_name][service] += service_impact

    return output_hydro_impacts

def group_offset_parcels_by_sshed(per_offset_data):
    """Return a dictionary mapping string serviceshed identifiers to lists of
    offset parcel IDs representing offset parcels that offset to that
    serviceshed."""
    output_sshed_data = {}
    for parcel_id, parcel_data in per_offset_data.iteritems():
        for sshed_name in parcel_data['municipalities'].keys():
            try:
                output_sshed_data[sshed_name].append(parcel_id)
            except KeyError:
                # if we haven't encountered this sshed name UNTIL NOW
                output_sshed_data[sshed_name] = [parcel_id]

    # now, loop through all the found ssheds and sort the parcel keys
    for sshed_name, offset_parcels in output_sshed_data.iteritems():
        output_sshed_data[sshed_name] = sorted(offset_parcels)

    return output_sshed_data





import sys
import os
import imp

import faulthandler
faulthandler.enable()  # enable for the ENTIRE adept package.
import natcap.versioner

__version__ = natcap.versioner.get_version('natcap.opal')

def test():
    """Autodiscover and run all tests in the adept package."""
    # use run() here, because it won't automatically quit the python interpreter
    # if we're running it within the interpreter.
    import natcap.opal.tests
    import unittest

    test_runner = unittest.TextTestRunner(verbosity=2)
    test_runner.run(natcap.opal.tests.test_suite())

def execute(args):
    import adept_core
    execute.__doc__ = adept_core.execute.__doc__
    adept_core.execute(args)

def is_frozen():
    if getattr(sys, 'frozen', False):
        # we are running in a |PyInstaller| bundle
        return True
        #basedir = os.path.dirname(sys.executable)
    else:
        return False
        #basedir = os.path.dirname(__file__)

def get_frozen_dir():
    return os.path.dirname(sys.executable)

def local_dir(file_path):
    if getattr(sys, 'frozen', False):
        # we are running in a |PyInstaller| bundle
        # get the folder containing the natcap folder.
        relpath = os.path.relpath(file_path, os.path.dirname(__file__))
        pkg_path = os.path.join('natcap', 'opal', relpath)
        return os.path.join(os.path.dirname(sys.executable), os.path.dirname(pkg_path))
    return os.path.dirname(file_path)





import unittest
import tempfile
import shutil
import os
import glob

import shapely.wkt
import shapely.ops
from shapely.geometry import Polygon, MultiPolygon, LineString, MultiLineString
import rtree
from osgeo import ogr

import natcap.opal.tests
from natcap.opal import preprocessing
from natcap.opal import utils

class PreprocessingTest(unittest.TestCase):
    def test_rm_shapefile(self):
        # Verify that all files are removed correctly.
        # first, build up a temp dir with mock files.
        temp_dir = tempfile.mkdtemp()
        def make_file(filename):
            uri = os.path.join(temp_dir, filename)
            new_file = open(uri, 'w')
            #new_file.write('')
            new_file.close()
            return uri

        files = [make_file("vector.%s" % ext) for ext in ['shp', 'shx', 'dbf']]
        keep = make_file('not_a_vector.csv') # should not be deleted

        preprocessing.rm_shapefile(files[0])  # remove vector.shp
        for vector_file in files:
            self.assertEqual(os.path.exists(vector_file), False)
        self.assertEqual(os.path.exists(keep), True)

        shutil.rmtree(temp_dir)

    def test_build_spatial_index(self):
        polygon_a = Polygon([(1, 1), (5, 1), (5, 5), (1, 5),
            (1, 1)])
        polygon_b = Polygon([(3, 3), (9, 3), (9, 9), (3, 9),
            (3, 3)])
        polygon_c = Polygon([(8, 1), (11, 1), (11, 4), (8, 4),
            (8, 1)])

        vector_uri = natcap.opal.tests.vector([polygon_a, polygon_b, polygon_c],
            natcap.opal.tests.COLOMBIA_SRS)

        spat_index, parcels = preprocessing.build_spatial_index(vector_uri)

        # verify that the number of parcels is correct
        self.assertEqual(len(parcels), 3)
        self.assertEqual(isinstance(spat_index, rtree.index.Index), True)

        natcap.opal.tests.cleanup(vector_uri)

    def test_split_multipolygons(self):
        base_coords = [(2, 2), (4, 2), (4, 4), (2, 4), (2, 2)]
        def translate(x, y):
            return [(a + x, b + y) for (a, b) in base_coords]

        polygon_a = Polygon(base_coords)
        polygon_b = Polygon(translate(4, 0))
        polygon_c = Polygon(translate(0, 4))
        polygon_d = Polygon(translate(4, 4))
        polygon_e = Polygon(translate(8, 2))

        multipoly_a = MultiPolygon([polygon_a, polygon_b])
        multipoly_b = MultiPolygon([polygon_c, polygon_d])

        multi_vec = natcap.opal.tests.vector([multipoly_a, multipoly_b, polygon_e],
            natcap.opal.tests.COLOMBIA_SRS, format='ESRI Shapefile')

        out_dir = tempfile.mkdtemp()
        out_vector_uri = os.path.join(out_dir, 'split.shp')
        preprocessing.split_multipolygons(multi_vec, out_vector_uri)

        # Check the split vector for contents.
        out_vector = ogr.Open(out_vector_uri)
        out_layer = out_vector.GetLayer()
        self.assertEqual(out_layer.GetFeatureCount(), 5)

        out_layer = None
        out_vector = None
        shutil.rmtree(out_dir)

    def test_split_multipolygons_line(self):
        """
        Split_multipolygons() should skip over features that are not polygons
        """
        base_coords = [(2, 2), (4, 2), (4, 4), (2, 4), (2, 2)]
        def translate(x, y):
            return [(a + x, b + y) for (a, b) in base_coords]

        linestring_a = LineString(base_coords)
        linestring_b = LineString(translate(4, 0))
        linestring_c = LineString(translate(0, 4))
        linestring_d = LineString(translate(4, 4))
        linestring_e = LineString(translate(8, 2))

        multilinestring_a = MultiLineString([linestring_a, linestring_b])
        multilinestring_b = MultiLineString([linestring_c, linestring_d])

        multi_vec = natcap.opal.tests.vector(
            [multilinestring_a, multilinestring_b, linestring_e],
            natcap.opal.tests.COLOMBIA_SRS, format='ESRI Shapefile')

        out_dir = tempfile.mkdtemp()
        out_vector_uri = os.path.join(out_dir, 'split.shp')
        self.assertRaises(ValueError,  preprocessing.split_multipolygons, multi_vec, out_vector_uri)

        shutil.rmtree(out_dir)


    def test_prepare_aoi(self):
        # create an impact sites vector.
        impact_a = Polygon([(5, 3), (10, 3), (10, 5), (5, 5), (5, 3)])
        impact_b = Polygon([(4, 9), (6, 9), (6, 11), (4, 11), (4, 9)])
        impacts_vector = natcap.opal.tests.vector([impact_a, impact_b],
            natcap.opal.tests.COLOMBIA_SRS, format='ESRI Shapefile')

        # create a sample hydrozubsones vector
        subzone_a = Polygon([(1, 1), (8, 1), (8, 7), (1, 7), (1, 1)])
        subzone_b = Polygon([(8, 1), (14, 1), (14, 8), (8, 8), (8, 1)])
        subzone_c = Polygon([(8, 8), (15, 8), (15, 14), (8, 14), (8, 8)])
        subzone_d = Polygon([(1, 7), (8, 7), (8, 14), (1, 14), (1, 7)])
        subzones_vector = natcap.opal.tests.vector([subzone_a, subzone_b, subzone_c,
            subzone_d], natcap.opal.tests.COLOMBIA_SRS, format='ESRI Shapefile')

        temp_dir = tempfile.mkdtemp()
        out_uri = os.path.join(temp_dir, 'aoi.shp')
        preprocessing.prepare_aoi(impacts_vector, subzones_vector, out_uri)

        out_vector = ogr.Open(out_uri)
        out_layer = out_vector.GetLayer()
        self.assertEqual(out_layer.GetFeatureCount(), 1)

        # take the union of the three subzone polygons that we expect.
        expected_geom = shapely.ops.cascaded_union([subzone_a, subzone_b,
            subzone_d])

        # get the wkt of the feature in out_layer.
        feature = out_layer.GetFeature(0)
        geometry = feature.GetGeometryRef()

        # convert to shapely geometry
        found_geom = shapely.wkt.loads(geometry.ExportToWkt())

        # assert that the areas are the same
        self.assertEqual(found_geom.area, expected_geom.area)

        # take the difference of the two polygons, assert that the resulting
        # area is 0.
        difference = expected_geom.difference(found_geom)
        self.assertEqual(difference.area, 0)

    def test_prepare_impact_sites(self):
        # build an impact sites vector.
        impact_a = Polygon([(5, 3), (10, 3), (10, 5), (5, 5), (5, 3)])
        impact_b = Polygon([(4, 9), (6, 9), (6, 11), (4, 11), (4, 9)])
        impacts_vector = natcap.opal.tests.vector([impact_a, impact_b],
            natcap.opal.tests.COLOMBIA_SRS, format='ESRI Shapefile')

        # build a hydrozones vector where two hydrozones intersect the same
        # impact polygon.
        subzone_a = Polygon([(1, 1), (8, 1), (8, 7), (1, 7), (1, 1)])
        subzone_b = Polygon([(8, 1), (14, 1), (14, 8), (8, 8), (8, 1)])
        subzone_c = Polygon([(8, 8), (15, 8), (15, 14), (8, 14), (8, 8)])
        subzone_d = Polygon([(1, 7), (8, 7), (8, 14), (1, 14), (1, 7)])
        fields = {
            'zone': str,
        }
        field_values = [
            {'zone': 'A'},
            {'zone': 'B'},
            {'zone': 'C'},
            {'zone': 'D'},
        ]
        subzones_vector = natcap.opal.tests.vector([subzone_a, subzone_b, subzone_c,
            subzone_d], natcap.opal.tests.COLOMBIA_SRS, format='ESRI Shapefile',
            fields=fields, features=field_values)

        temp_dir = tempfile.mkdtemp()

        preprocessing.prepare_impact_sites(impacts_vector, subzones_vector,
            temp_dir)

        # prepare_impact_sites creates a folder with output vectors.  There
        # should be 3 of them.
        geom_a = Polygon([(5, 3), (8, 3), (8, 5), (5, 5), (5, 3)])
        geom_b = Polygon([(8, 3), (10, 3), (10, 5), (8, 5), (8, 3)])
        geom_d = impact_b  # this impact should not have changed.
        for zone, impact_geom in [('A', geom_a), ('B', geom_b), ('D', geom_d)]:
            impact_filename = os.path.join(temp_dir,
                'impacts_%s.shp' % zone)
            self.assertTrue(os.path.exists(impact_filename))

            impact_vector = ogr.Open(impact_filename)
            impact_layer = impact_vector.GetLayer()
            self.assertEqual(impact_layer.GetFeatureCount(), 1)

            feature = impact_layer.GetFeature(0)
            shapely_feature = utils.build_shapely_polygon(feature)
            self.assertEqual(impact_geom.union(shapely_feature).area,
                impact_geom.area)
            self.assertEqual(impact_geom.difference(shapely_feature).area, 0.0)

        shutil.rmtree(temp_dir)

    def test_union_by_attribute(self):
        # build a sample vector
        polygon_a = Polygon([(1, 1), (2, 1), (2, 2), (1, 2), (1, 1)])
        polygon_b = Polygon([(2, 1), (3, 1), (3, 2), (2, 2), (2, 1)])
        polygon_c = Polygon([(3, 1), (4, 1), (4, 2), (3, 2), (3, 1)])

        fields = {
            'zone': str
        }
        field_values = [
            {'zone': 'first'},
            {'zone': 'first'},
            {'zone': 'second'},
        ]
        sample_vector = natcap.opal.tests.vector([polygon_a, polygon_b, polygon_c],
            natcap.opal.tests.COLOMBIA_SRS, fields, field_values)

        temp_dir = tempfile.mkdtemp()
        out_file = os.path.join(temp_dir, 'union_by_attribute.shp')
        preprocessing.union_by_attribute(sample_vector, 'zone', out_file)

        out_vector = ogr.Open(out_file)
        out_layer = out_vector.GetLayer()
        self.assertEqual(out_layer.GetFeatureCount(), 2)

        expected_geom = Polygon([(1, 1), (3, 1), (3, 2), (1, 2), (1, 1)])
        first_ogr_feature = out_layer.GetFeature(1)
        first_feature = utils.build_shapely_polygon(first_ogr_feature)
        self.assertEqual(expected_geom.area,
            expected_geom.union(first_feature).area)
        self.assertEqual(expected_geom.difference(first_feature).area, 0.0)

        # assert that the field was created in the layer and that the field
        # value was set for the first field..
        fieldnames = map(lambda x: x.GetName(), out_layer.schema)
        self.assertTrue('zone' in fieldnames)
        self.assertEqual(first_ogr_feature.GetField('zone'), 'first')

        shutil.rmtree(temp_dir)


    def test_locate_intersecting_polygons(self):
        polygon_a = Polygon([(1, 1), (5, 1), (5, 5), (1, 5),
            (1, 1)])
        polygon_b = Polygon([(3, 3), (9, 3), (9, 9), (3, 9),
            (3, 3)])
        polygon_c = Polygon([(8, 1), (11, 1), (11, 4), (8, 4),
            (8, 1)])

        vector_uri = natcap.opal.tests.vector([polygon_a, polygon_b, polygon_c],
            natcap.opal.tests.COLOMBIA_SRS)

        comparison_polygon = Polygon([(0, 0), (4, 0), (4, 4), (4, 0), (0, 0)])
        comp_vector_uri = natcap.opal.tests.vector([comparison_polygon],
                                                   natcap.opal.tests.COLOMBIA_SRS)

        out_dir = tempfile.mkdtemp()
        out_vector_uri = os.path.join(out_dir, 'out_vector.shp')
        preprocessing.locate_intersecting_polygons(vector_uri, comp_vector_uri,
                                                   out_vector_uri)

        # TODO: use this test (or even better, a test on real sample data)
        # to compare OGR layer operations against per-polygon operations.





"""Adept_core contains the core logic for the natcap.opal permitting tool."""
# -*- coding: utf-8 -*-

import json
import glob
import logging
import os
import shutil
import tempfile
from types import BooleanType
from types import StringType
import time

from osgeo import ogr
import pygeoprocessing

from natcap.invest import reporting
import natcap.opal
import natcap.opal.i18n
import preprocessing
import offsets
import static_maps
import reporting as opal_reporting
import analysis
import utils

logging.basicConfig(format='%(asctime)s %(name)-18s %(levelname)-8s \
     %(message)s', level=logging.DEBUG, datefmt='%m/%d/%Y %H:%M:%S ')

LOGGER = logging.getLogger('natcap.opal')

SOURCE_DATA = os.path.join(os.getcwd(), 'data', 'colombia_tool_data')
REPORT_DATA = os.path.join(natcap.opal.local_dir(__file__), 'report_data')
_ = natcap.opal.i18n.language.ugettext
LABEL_PROTECTION = 'Protection'
LABEL_RESTORATION = 'Restoration'
DIST_OPAL = 'opal'
DIST_MAFE = 'mafe-t'

class InvalidImpactsVector(Exception): pass
class InvalidInput(Exception): pass

def execute(args):
    """The main execution function for natcap.opal.
        args - a python dictionary containing the following elements:
            'workspace_dir' - A uri to the output workspace.
            'project_footprint_uri' - a URI to a shapefile of the project
                footprint
            'impact_type' - "Road/Mine" or "Bare ground/Paved".
            'area_of_influence_uri' - a uri to a shapefile that defines the
                area of influcence of the impact site
            'ecosystems_map_uri' - a URI to an OGR vector with all the
                ecosystems polygons.
            'offset_parcels' - a URI to an OGR vector with parcels that can be
                selected as offset parcels.  May be the same URI as
                args['ecosystems_map_uri'] in cases where args['offset_scheme']
                is 'Protection'.
            'threat_map' - ?
            'species_richness_map' - ?
            'search_areas_uri' - a URI to an OGR vector containing search area.
                TODO: rename this to be hydrosubzones.
            'avoidance_areas' - a URI to an OGR vector of areas within which
                offset parcels may not be selected.
            'conservation_portfolio' - a URI to an OGR vector of priority areas
                for conservation.  Regions defined in the conservation
                portfolio will be selected over other offset parcels within the
                same search area.
            'previously_granted_impacts' - ?
            'previously_selected_offsets' - an OGR vector with previously
                selected offset parcels.  These parcels may not be selected
                again.
            'municipalities_uri' - A URI to a vector of municipalities. If not
                provided, the tool will try to use the internal municipalities
                vector.  If neither is provided nor available, an exception
                will be raised.
            'data_dir' - a URI to a folder that contains a folder called
                'data', which itself contains the necessary colombia tool data.
            'sediment_static_maps' - a uri to a folder that contains the three
                static maps generated. (optional)
            'nutrient_static_maps' - a uri to a folder that contains the three
                static maps generated. (optional)
            'carbon_static_maps' - a uri to a folder that contains the three
                static maps generated. (optional)
            'custom_static_maps' - a uri to a folder that contains the three
                static maps generated. (optional)
            'custom_servicesheds' - a string, either 'hydrological' or
                'global'.  Required if custom_static_maps argument is provided.
                If 'hydrological', the custom_static_maps folder must also
                contain appropriate percent-to-stream rasters, following the
                same naming convention as the other routed static data folders.
            'do_carbon' - Boolean.  Optional.  Whether to include carbon in
                calculations.  Defaults to True if not specified.
            'do_nutrient' - Boolean.  Optional.  Whether to include nitrogen
                in calculations. Defaults to True if not specified.
            'do_sediment' - Boolean.  Optional.  Whether to include sediment
                in calculations. Defaults to True if not specified.
            'offset_scheme' - a number.  The number must be one of [0, 1, 2],
                and indicates the following offset selection schemes:
                    0 - Ecosystem services only will be considered when
                        selecting offset parcels.
                    1 - Biodiversity only will be considered when selecting
                        offset parcels.
                    2 - Both biodiversity and ecosystem services will be
                        considered when selecting offset parcels.
                    This input is optional.  When not provided, defaults to 1.
            'carbon_mitigation_ratio' - (optional) The numeric factor by which
                carbon impact quantities will be multiplied to calculate the
                required offset quantitiy in the report.  This defaults to 1.0
                if not provided by the user.
            'nutrient_mitigation_ratio' - (optional) The numeric factor by which
                nutrient impact quantities will be multiplied to calculate the
                required offset quantitiy in the report.  This defaults to 1.0
                if not provided by the user.
            'sediment_mitigation_ratio' - (optional) The numeric factor by which
                sediment impact quantities will be multiplied to calculate the
                required offset quantitiy in the report.  This defaults to 1.0
                if not provided by the user.
            'custom_mitigation_ratio' - (optional) The numeric factor by which
                custom ES impact quantities will be multiplied to calculate the
                required offset quantitiy in the report.  This defaults to 1.0
                if not provided by the user.
            'prop_offset' - (optional) the proportion of (impacts * mitigation
                ratio) that the set of recommended parcels should offset.
                Defaults to 1.0 if not user-defined.
            'distribution' - (optional) The distribution of the tool we're
                running.  Either 'mafe-t' or 'opal'.  Defaults to 'mafe-t' if
                not provided.

        Returns nothing."""

    LOGGER.debug('Current language: "%s"', natcap.opal.i18n.language.current_lang)
    # build a list of possible places to look for the ascii art text file in
    # order of priority.
    possible_dirs = []
    if natcap.opal.is_frozen():
        possible_dirs += [natcap.opal.get_frozen_dir()]
    possible_dirs += [os.path.dirname(__file__), os.getcwd()]
    ascii_art_uri = ''
    for dirname in possible_dirs:
        ascii_art_uri = os.path.join(dirname, 'ASCII_NAME.txt')
        if os.path.exists(ascii_art_uri):
            break

    if os.path.exists(ascii_art_uri):
        with open(ascii_art_uri) as ascii_art:
            for line in ascii_art:
                LOGGER.debug(r'%s', line.rstrip())
    else:
        LOGGER.debug(r'   _____       .___             __    ')
        LOGGER.debug(r'  /  _  \    __| _/____ _______/  |_  ')
        LOGGER.debug(r' /  /_\  \  / __ |/ __ \\____ \   __\ ')
        LOGGER.debug(r'/    |    \/ /_/ \  ___/|  |_> >  |   ')
        LOGGER.debug(r'\____|__  /\____ |\___  >   __/|__|   ')
        LOGGER.debug(r'        \/      \/    \/|__|          ')
        LOGGER.debug(r'                                      ')

    common_data = static_maps.get_common_data_json(args['data_dir'])

    dirs = {
        'workspace': os.path.abspath(args['workspace_dir']),
        'intermediate': os.path.join(args['workspace_dir'], 'intermediate'),
        'impact_sites': os.path.join(args['workspace_dir'], 'intermediate',
            'impact_sites'),
        'temp': os.path.join(args['workspace_dir'], 'temp'),
        'results': os.path.join(args['workspace_dir'], 'results'),
    }
    LOGGER.debug('workspace directories: %s', json.dumps(dirs, indent=4,
        sort_keys=True))
    old_temp_dir = tempfile.tempdir
    tempfile.temp_dir = dirs['temp']

    try:
        _ecosystems = args['ecosystems_map_uri']
    except KeyError:
        _ecosystems = common_data['ecosystems']

    try:
        # allow the user to not include offset_parcels.  This would be default
        # behavior in MAFE-T
        _offset_parcels = args['offset_parcels']
    except KeyError:
        _offset_parcels = _ecosystems

    files = {
        'active_hydrozones': os.path.join(dirs['intermediate'],
            'active_hydrozones.shp'),
        'max_search_area': os.path.join(dirs['intermediate'],
            'search_area.shp'),
        'ecosystems': _ecosystems,
        'offset_parcels': _offset_parcels,
        'prep_offset_sites': os.path.join(dirs['temp'],
            'tmp_offset_sites.shp'),
        'prep_natural_parcels': os.path.join(dirs['intermediate'],
            'prepared_ecosystems.shp'),
        'global_impacted_subzones': os.path.join(dirs['intermediate'],
            'impacted_subzones.shp'),
        'union_of_subzones': os.path.join(dirs['intermediate'],
            'union_impacted_subzones.shp'),
        'buffered_subzones': os.path.join(dirs['intermediate'],
            'buffered_subzones.shp'),
        'base_report': os.path.join(dirs['results'], 'index.html'),
        'impacted_sb3': os.path.join(dirs['intermediate'],
                                     'impacted_softboundary3.shp')
    }

    pygeoprocessing.create_directories(dirs.values())

    try:
        if args['offset_scheme'] not in [0, 1, 2]:
            raise InvalidInput(('Offset scheme %s is invalid.  Must be ',
                ' in [0, 1, 2]') % args['offset_scheme'])
    except KeyError:
        # raised when the offset_scheme key is not in args.
        # When this happens, assume the default behavior, to use biodiversity
        # as an offset scheme.
        args['offset_scheme'] = offsets.OFFSET_SCHEME_BIODIV

    # a list of all the ecosystem services we care about right now.  Only using
    # the custom static map right now as a proxy for sediment.
    # TODO: if user provided custom static map, add a 'custom' service.
    # TODO: Add a static protection map here.
    LOGGER.debug("User-defined impact type: %s", args['impact_type'])
    assert args['impact_type'] in ['Road/Paved', 'Mine/Bare'], ('Invalid '
        'impact type %s' % args['impact_type'])

    if args['impact_type'] == 'Road/Paved':
        impact_key = 'paved'
    else:
        impact_key = 'bare'
    LOGGER.debug('Using impact type %s', impact_key)

    LOGGER.debug('Checking desired services to consider')
    desired_services = []
    for service in ['carbon', 'nutrient', 'sediment']:
        key = 'do_%s' % service
        try:
            do_service = args[key]
        except KeyError:
            do_service = True
        if do_service:
            desired_services.append(service)

    if len(desired_services) == 0:
        raise RuntimeError('You must select at least one service')
    LOGGER.debug('Considering services %s', desired_services)

    if 'custom_static_maps' in args:
        LOGGER.debug('User specified that custom ES should be considered')
        desired_services.append('custom')
        assert args['custom_servicesheds'] in ['hydrological', 'global']
        custom_servicesheds = args['custom_servicesheds']
    else:
        custom_servicesheds = None

    impact_services = []
    protection_services = []
    for es_name in desired_services:
        args_key = '%s_static_maps' % es_name

        use_percent_to_stream = False
        if es_name in ['nutrient']:
            use_percent_to_stream = True
        elif es_name == 'custom':
            LOGGER.debug('Checking for custom ES inputs.')
            # We only want to use percent_to_stream if we're using hydrological
            # servicesheds.
            # this might raise a keyError if the user doesn't provide the right
            # inputs.  They should expect an error in this case anyways.
            if args['custom_servicesheds'] == 'hydrological':
                LOGGER.debug('Custom ES inputs are hydrological')
                use_percent_to_stream = True

        try:
            if args['offset_type'] == LABEL_PROTECTION:
                future_base = 'protect'
            elif args['offset_type'] == LABEL_RESTORATION:
                future_base = 'restore'
            else:
                raise InvalidInput(('Offset type is invalid.  Must be one of ',
                    '"Protection" or "Restoration", "%s" found' % future_base))

            static_maps_dir = args[args_key]
            LOGGER.debug('User defined custom static maps for %s', es_name)
            fmt_string = '%s_%s_static_map.tif'
            impact_filename = fmt_string % (es_name, impact_key)
            protection_filename = fmt_string % (es_name, future_base)

            if use_percent_to_stream:
                pts_fmt_string = '%s_%s_pts.tif'
                impact_percent_to_stream = os.path.join(static_maps_dir,
                    pts_fmt_string % (es_name, impact_key))
                protect_percent_to_stream = os.path.join(static_maps_dir,
                    pts_fmt_string % (es_name, future_base))
            else:
                impact_percent_to_stream = None
                protect_percent_to_stream = None

            static_impact_map = os.path.join(static_maps_dir, impact_filename)
            static_protection_map = os.path.join(static_maps_dir,
                protection_filename)
            LOGGER.debug('Using custom static impact map %s',
                static_impact_map)
            LOGGER.debug('Using custom static protection map %s',
                static_protection_map)
        except KeyError:
            # If The user defined custom static map values, we want to
            # consider them.  If not, we want to skip them.
            if es_name == 'custom':
                continue

            static_impact_map = common_data['static_maps'][impact_key][es_name]
            static_protection_map = common_data['static_maps']['protection'][es_name]

            if use_percent_to_stream:
                impact_percent_to_stream = common_data['percent_to_stream'][es_name][impact_key]
                protect_percent_to_stream = common_data['percent_to_stream'][es_name]['protection']
            else:
                impact_percent_to_stream = None
                protect_percent_to_stream = None

            LOGGER.debug('No custom static maps defined for %s', es_name)
            LOGGER.debug('Using default static impact map %s',
                static_impact_map)
            LOGGER.debug('Using default static future scenario map %s',
                static_protection_map)


        static_data = {
            'static_impact': static_impact_map,
            'static_protection': static_protection_map,
            'pts_impact': impact_percent_to_stream,
            'pts_protection': protect_percent_to_stream,
        }

        protection_services.append((es_name, static_data))

    # verify that the correct maps exist on disk.  If they don't exist, raise
    # an error.  They might not have installed static maps with the Colombia
    # installer.
    for es_name, static_data in protection_services:
        LOGGER.debug('Verifying static maps exist for %s', es_name)

        files_to_check = [
            ('static_impact', 'static impact'),
            ('static_protection', 'static protection'),
            ('pts_impact', 'impact percent-to-stream'),
            ('pts_protection', 'protection percent-to-stream'),
        ]

        for raster_key, label in files_to_check:
            file_uri = static_data[raster_key]
            if file_uri is None:
                continue

            if not os.path.exists(file_uri):
                file_uri = os.path.abspath(file_uri)
                LOGGER.critical('Expecting %s map at %s, but not found',
                    label, file_uri)
                raise IOError('File not found: %s' % file_uri)

    LOGGER.debug('Considering services: %s', [r[0] for r in protection_services])

    # get a list of all static maps for all services.
    all_static_maps = []
    for map_list in [s[1].values() for s in protection_services]:
        for map_uri in map_list:
            if map_uri is not None:
                all_static_maps.append(map_uri)
    if not preprocessing.impacts_in_static_maps(all_static_maps,
            args['project_footprint_uri']):
        raise InvalidImpactsVector(('Some polygon(s) do not intersect static '
            'maps.  Correct this and re-run the tool.'))

    # Check distribution if it's provided.  Otherwise, assume we're in OPAL.
    if 'distribution' in args:
        if args['distribution'] not in [DIST_OPAL, DIST_MAFE]:
            raise RuntimeError('Distribution must be one of %s, %s found' %
                ([DIST_OPAL, DIST_MAFE], args['distribution']))
    else:
        args['distribution'] = DIST_OPAL
    LOGGER.debug('Using distribution %s' % args['distribution'])

    try:
        hydro_subzones = args['search_areas_uri']
        LOGGER.debug('Using user-provided hydro subzones: %s', hydro_subzones)

        hydrozones = os.path.join(dirs['intermediate'], 'hydrozones.shp')
        # build the hydrozones out of the hydrosubzones by zone attribute
        contained_subzones = preprocessing.union_by_attribute(hydro_subzones,
            'zone', hydrozones)
    except KeyError:
        hydro_subzones = common_data['hydrosubzones']
        hydrozones = common_data['hydrozones']
        LOGGER.debug('Using default hydro subzones: %s', hydro_subzones)
        LOGGER.debug('Using default hydrozones: %s', hydrozones)

    try:
        municipalities = args['municipalities_uri']
        LOGGER.debug('Using user-provided municipalities: %s', municipalities)
    except KeyError:
        municipalities = common_data['municipalities']
        LOGGER.debug('Using default municipalities: %s', municipalities)
    finally:
        # Raise error about required municipalities only if the file can't be
        # found.
        if args['distribution'] == DIST_OPAL:
            if not os.path.exists(municipalities):
                municipalities = None
            else:
                # if the user prvided a softboundary3 (municipalities), save off a vector
                # of all the municipalities that intersect the impact sites.
                preprocessing.locate_intersecting_polygons(
                    municipalities, args['project_footprint_uri'],
                    files['impacted_sb3'])
        else:
            raise RuntimeError(("Municipalities vector is required, "
                "but was not provided or could not be found"))

    try:
        area_of_influence = args['area_of_influence_uri']
        LOGGER.debug('Using user-provided AOI')
    except KeyError:
        LOGGER.debug('Building AOI from hydro subzones')
        area_of_influence = os.path.join(dirs['intermediate'], 'aoi_computed.shp')
        preprocessing.prepare_aoi(args['project_footprint_uri'],
            hydro_subzones, area_of_influence)

    LOGGER.info('Determining active hydrozone')
    preprocessing.locate_intersecting_polygons(hydrozones,
        args['project_footprint_uri'], files['active_hydrozones'])

    LOGGER.info('Preparing impact sites')
    impact_sites_list = preprocessing.prepare_impact_sites(args['project_footprint_uri'],
        hydrozones, dirs['impact_sites'])

    LOGGER.info('Finding the union of active hydrozones and the AOI')
    preprocessing.union_of_vectors([files['active_hydrozones'],
        area_of_influence], files['max_search_area'])

    if 'include_lci' not in args:
        args['include_lci'] = True
    else:
        assert type(args['include_lci']) is BooleanType, ('Key "include_lci"'
            ' must be a boolean.  %s found' % type(args['include_lci']))

    LOGGER.info('Preparing offset sites')
    preprocessing.prepare_offset_parcels(files['offset_parcels'],
        files['active_hydrozones'], files['prep_offset_sites'],
        include_lci=args['include_lci'])

    # if the natural and offset parcels are found in different vectors, then
    # we'll need to prepare a subset of the natural parcels as well.
    # If natural and offset parcels are found in the same vector, then we can
    # use the prepared offset parcels vector for the prepared natural parcels.
    _abspath = lambda p: os.path.abspath(p)
    if _abspath(files['offset_parcels']) != _abspath(files['ecosystems']):
        LOGGER.info('Preparing impacted natural parcels')

        # Build a vector of the hydro subzones that intersect the impact sites.
        preprocessing.locate_intersecting_polygons(args['search_areas_uri'],
            args['project_footprint_uri'], files['global_impacted_subzones'])
        preprocessing.union_of_vectors([files['global_impacted_subzones']],
            files['union_of_subzones'])

        # when buffering, include at least 1000 extra distance units of
        # parcels, just in case there are any parcels near the boundary of the
        # hydrozones that need to be included in the LCI calculations.
        preprocessing.buffer_vector(files['union_of_subzones'], 1000,
            files['buffered_subzones'])

        preprocessing.prepare_offset_parcels(files['ecosystems'],
            files['buffered_subzones'], files['prep_natural_parcels'],
            include_lci=args['include_lci'])
    else:
        LOGGER.debug('Offset parcels are natural')
        files['prep_natural_parcels'] = files['prep_offset_sites']


    try:
        servicesheds = args['servicesheds_uri']
        LOGGER.debug('User provided custom servicesheds: %s', servicesheds)
    except KeyError:
        servicesheds = common_data['servicesheds']
        LOGGER.debug('Using default servicesheds: %s', servicesheds)

    # Start looping through all of the impacted hydrozones
    for impact_sites_data in impact_sites_list:
        if type(impact_sites_data['name']) is StringType:
            recode = lambda s: s.decode('utf-8')
        else:
            recode = lambda s: s.encode('utf-8')
        impact_sites_data['name'] = recode(impact_sites_data['name'])

        LOGGER.debug('Processing impacts for hydrozone %s',
            impact_sites_data['name'])
        _clean_hydrozone_name = impact_sites_data['name'].lower().replace(' ', '_')

        hzone_dir = os.path.join(dirs['results'], _clean_hydrozone_name)
        hzone_dev = os.path.join(hzone_dir, '_dev')
        hzone_static_maps = os.path.join(hzone_dir, 'static_data')
        pygeoprocessing.create_directories([hzone_dir, hzone_dev,
            hzone_static_maps])

        hzone_paths = {
            'offset_sites': os.path.join(hzone_dir, 'offset_sites_available.shp'),
            'all_offsets': os.path.join(hzone_dir, 'potential_offsets_in_zone.shp'),
            'all_natural_parcels': os.path.join(hzone_dir, 'natural_parcels_in_zone.shp'),
            'impact_sites': impact_sites_data['uri'],
            'bio_impacts': os.path.join(hzone_dev, 'bio_impacts.json'),
            'selected_offsets': os.path.join(hzone_dir, 'offset_sites_filtered.shp'),
            'impacted_muni': os.path.join(hzone_dir, 'impacted_softboundary2.shp'),
            'servicesheds': os.path.join(hzone_dir, 'servicesheds_in_zone.shp'),
            'hydrozone': os.path.join(hzone_dir, 'impacted_zone.shp'),
            'hydrosubzones': os.path.join(hzone_dir, 'impacted_subzones.shp'),
            'parcel_info': os.path.join(hzone_dev, 'selected_parcels.json'),
        }

        LOGGER.info('Locating the current hydrozone polygon')
        preprocessing.locate_intersecting_polygons(files['active_hydrozones'],
            hzone_paths['impact_sites'], hzone_paths['hydrozone'])

        # TODO: use a split_vector function?
        LOGGER.info('Locating servicesheds intersecting this hydrozone')
        preprocessing.locate_intersecting_polygons(servicesheds,
            hzone_paths['hydrozone'], hzone_paths['servicesheds'])

        LOGGER.info('Locating offset parcels available for this hydrozone')
        preprocessing.locate_intersecting_polygons(files['prep_offset_sites'],
            hzone_paths['hydrozone'], hzone_paths['all_offsets'], clip=True)

        LOGGER.info('Locating natural parcels available for this hydrozone')
        preprocessing.locate_intersecting_polygons(files['prep_natural_parcels'],
            hzone_paths['hydrozone'], hzone_paths['all_natural_parcels'],
            clip=True)

        LOGGER.info('Cutting out the impact sites from available offset parcels.')
        preprocessing.subtract_vectors(hzone_paths['all_offsets'],
            hzone_paths['impact_sites'], hzone_paths['offset_sites'])

        LOGGER.info('Aggregating services provided by parcels')
        service_impacts = {}
        for service_name, static_data in protection_services:
            LOGGER.debug('Aggregating stats for service %s', service_name)
            LOGGER.debug('Aggregating %s stats for offset sites', service_name)
            _ = analysis.aggregate_stats(static_data['static_protection'],
                hzone_paths['offset_sites'], 'FID', service_name,
                static_data['pts_protection'])
            LOGGER.debug('Aggregating %s stats for impact sites', service_name)
            service_impacts[service_name] = analysis.aggregate_stats(
                static_data['static_impact'], hzone_paths['impact_sites'], 'FID',
                service_name, static_data['pts_impact'])

        for args_key, col_name in [('threat_map', 'Threat'), ('richness_map',
            'Richness')]:
            if args_key not in args:
                continue
            LOGGER.debug('Found %s map in args.  Calculating stats.',
                col_name.lower())
            _ = analysis.aggregate_stats(args[args_key],
                hzone_paths['impact_sites'], 'FID', col_name)
            _ = analysis.aggregate_stats(args[args_key],
                hzone_paths['offset_sites'], 'FID', col_name)
            _ = analysis.aggregate_stats(args[args_key],
                hzone_paths['all_offsets'], 'FID', col_name)

    #    LOGGER.info('Selecting offset parcels')
    #    offset_dir = os.path.join(output_dir, 'offsets')
    #    offset_info = offsets.select_offsets(offset_sites, impact_sites, offset_dir)
    #
    #    # find the impact sum
    #    total_impacts = {
    #        'sediment': sum(map(lambda r: r['services']['sediment']['impact'],
    #            offset_info.values()))
    #    }
    #
    #    LOGGER.debug(json.dumps(offset_info, sort_keys=True, indent=4))

        # get the proper impacts from the returned service impacts from the
        # above analysis step.
        _get_impacts = lambda k: sum(service_impacts[k].values())
        total_impacts = dict((service, _get_impacts(service)) for service in
            desired_services)
#        total_impacts = {
#            'sediment': _get_impacts('sediment'),
#            'nutrient': _get_impacts('nutrient'),
#            'carbon': _get_impacts('carbon'),
#        }
        if 'custom' in service_impacts:
            total_impacts['custom'] = _get_impacts('custom')

        LOGGER.debug('Total impacts: %s', total_impacts)

        # TODO: Write these values to JSON in the calculat_bio_impact function
        biodiversity_impact = analysis.calculate_biodiversity_impact(
            hzone_paths['impact_sites'], hzone_paths['all_natural_parcels'])
        json.dump(biodiversity_impact, open(hzone_paths['bio_impacts'], 'w'), sort_keys=True,
            indent=4)

        LOGGER.info(biodiversity_impact)

        # Build a vector of the current hydro subzones.
        preprocessing.locate_intersecting_polygons(hydro_subzones,
            hzone_paths['impact_sites'], hzone_paths['hydrosubzones'])

        comparison_vectors = {
            'AOI': area_of_influence,
            'Subzone': hzone_paths['hydrosubzones'],
        }

        try:
            comparison_vectors['Conservation Portfolio'] = \
                args['conservation_portfolio']
        except KeyError:
            LOGGER.debug('Conservation portfolio not in args, skipping.')

        if municipalities is not None:
            # Get all the municipalities that intersect with the impact vector.
            preprocessing.locate_intersecting_polygons(municipalities,
                hzone_paths['impact_sites'], hzone_paths['impacted_muni'])
            comparison_vectors['City'] = hzone_paths['impacted_muni']

        #biodiversity_impact contains ONLY the biodiversity impacts.
        _, recommended_parcels = offsets._select_offsets(
            hzone_paths['offset_sites'],
            hzone_paths['impact_sites'], biodiversity_impact,
            hzone_paths['selected_offsets'], hzone_paths['parcel_info'],
            comparison_vectors, total_impacts.keys(),
            offset_scheme=args['offset_scheme'])

        # take the parcels selected in _select_offsets and calculate the
        # percent_overlap
        temp_municipalities = os.path.join(hzone_dir,
            'servicesheds_with_offset_parcels.shp')
        per_offset_data = analysis.percent_overlap(
            hzone_paths['selected_offsets'], hzone_paths['servicesheds'],
            temp_municipalities, 'pop_center')
        json.dump(per_offset_data, open(os.path.join(hzone_dev,
            'per_offset_data.json'), 'w'), indent=4, sort_keys=True)
        opal_reporting.write_per_offset_csv(per_offset_data,
            os.path.join(hzone_dir, 'offset_benefits_per_serviceshed.csv'))


        #########################
        # PARCEL RECOMMENDATION #
        #########################
        # parcel data contains information for us to drive the offset value per
        # serviceshed as well.
        parcel_data = offsets.translate_parcel_data(per_offset_data)
        json.dump(parcel_data, open(os.path.join(hzone_dev,
            'translated_parcel_data.json'), 'w'), indent=4, sort_keys=True)

        # 2 of the offset schemes use ES.  If we're using ES, include ES parcel
        # analysis and offset recommendation.
        if args['offset_scheme'] != offsets.OFFSET_SCHEME_BIODIV:
            temp_municipalities_2 = os.path.join(dirs['temp'],
                'municipalities_intersecting_impacts.shp')
            per_impact_data = analysis.percent_overlap(
                hzone_paths['impact_sites'], hzone_paths['servicesheds'],
                temp_municipalities_2, 'pop_center')
            json.dump(per_impact_data, open(os.path.join(hzone_dev,
                'per_impact_data.json'), 'w'), indent=4, sort_keys=True)

            # collect ES impacts from per_impact_data
            es_hydro_requirements = offsets.translate_es_impacts(per_impact_data)
            json.dump(es_hydro_requirements, open(os.path.join(hzone_dev,
                'es_hydro_requirements.json'), 'w'), indent=4, sort_keys=True)

            # group offset parcels by serviceshed and merge them into the
            # es_hydro_requirements
            offset_parcels_by_sshed = offsets.group_offset_parcels_by_sshed(
                per_offset_data)
            empty_services = dict((k, 0) for k in es_hydro_requirements.values()[0])
            for sshed_name, sshed_offsets in offset_parcels_by_sshed.iteritems():
                try:
                    es_hydro_requirements[sshed_name]['parcels'] = sshed_offsets
                except KeyError:
                    # when there are no impacts for this serviceshed (but there are
                    # offsets), then we can skip the serviceshed.
                    continue
            json.dump({'offset_parcels_by_sshed': offset_parcels_by_sshed,
                'es_hydro_requirements': es_hydro_requirements},
                open(os.path.join(hzone_dev,
                'es_hydro_requirements_complete.json'), 'w'), indent=4,
                sort_keys=True)
        else:
            es_hydro_requirements = None

        # biodiversity requirements stored in the `biodiversity_impact`
        # variable.
        # hydro_es_req - this would need to be determined from the
        # per_impact_data
        if 'prop_offset' not in args:
            LOGGER.debug('Defaulting to prop_offset=1.0')
            args['prop_offset'] = 1.0
        recommended_parcels = offsets.select_set_multifactor(parcel_data,
            biodiversity_impact, es_hydro_requirements,
            proportion_offset=args['prop_offset'])
        json.dump(recommended_parcels, open(os.path.join(hzone_dev,
            'recommended_parcels.json'), 'w'), indent=4, sort_keys=True)

        # check to see if the user provided avoidance areas.  If so, check to
        # see if any of the impact sites intersect the provided avoidance
        # areas.
        try:
            impacts_disallowed = analysis.vectors_intersect(
                hzone_paths['impact_sites'], args['avoidance_areas'])
        except KeyError:
            impacts_disallowed = False

        service_mitigation_ratios = {}
        for service in total_impacts.keys():
            mit_ratio_key = '%s_mitigation_ratio' % service
            try:
                mit_ratio = float(args[mit_ratio_key])
            except KeyError:
                mit_ratio = 1.0
            service_mitigation_ratios[service] = mit_ratio

        # We only want to include the AOI in the report if the user actually
        # provided an AOI (so we did not compute it for them), and we're
        # running OPAL.
        include_aoi = True
        if 'area_of_influence_uri' not in args:
            if args['distribution'] == DIST_OPAL:
                include_aoi = False
        LOGGER.debug('Include AOI column: %s', include_aoi)

        # We only want to include the subzones column in the parcels CSV iff
        # there is only 1 subzone for this zone.
        include_subzone = impact_sites_data['name'] in contained_subzones
        LOGGER.debug('Name: %s', impact_sites_data['name'])
        LOGGER.debug('Contained subzones: %s', contained_subzones)
        LOGGER.debug('Include subzone column: %s', include_subzone)

        LOGGER.info("Building output report")
        build_report(hzone_paths['servicesheds'], biodiversity_impact,
            hzone_paths['selected_offsets'],
            args['project_footprint_uri'], total_impacts,
            args['impact_type'], hzone_dir, hzone_paths['impact_sites'],
            'pop_center', '%s_report.html' % _clean_hydrozone_name,
            hzone_paths['all_natural_parcels'], impacts_disallowed,
            suggested_parcels=recommended_parcels,
            custom_es_servicesheds=custom_servicesheds,
            dev_dir = hzone_dev, service_mitrat=service_mitigation_ratios,
            per_offset_data=per_offset_data, prop_offset=args['prop_offset'],
            distribution=args['distribution'], include_aoi_column=include_aoi,
            include_subzone_column=include_subzone,
            hzone_name=impact_sites_data['name'])

        # copy static maps to the workspace.
        LOGGER.info('Clipping static data to the hydrozone for reference')
        for service_name, static_data in protection_services:
            for impact_raster_key in ['static_impact', 'static_protection']:
                src_file = static_data[impact_raster_key]
                dest_filename = os.path.join(hzone_static_maps,
                    '%s_%s.tif' % (service_name, impact_raster_key))
                LOGGER.info('Clipping %s:%s', impact_raster_key, src_file)
                try:
                    static_maps.clip_static_map(src_file, hzone_paths['hydrozone'],
                        dest_filename)
                except Exception as error:
                    LOGGER.error('Could not clip static map: %s', error)


    # return tempfile.tempdir to its original state.
    tempfile.tempdir = old_temp_dir
    shutil.rmtree(dirs['temp'])

    write_results_index(dirs['results'], files['base_report'],
                         distribution=args['distribution'])

def write_results_index(results_dir, out_file, distribution):
    """
    Write an HTML page to the results folder with links to per-zone analyses.

    This HTML page is extremely simple relative to the interactivity contained
    in the analysis pages.  It's just a UL with some helptext, where each LI
    links to the given hydrozones analysis.

    Parameters:
        results_dir (string): A string filepath to the folder containing each
            of the per-hydrozone results directories.
        out_file (string): The path to the file to which the HTML file will be
            written.
        distribution (string): The name of the distribution to write to the
            file.

    Returns:
        None
    """
    zone_glob = os.path.join(results_dir, '*')
    zone_directory_basenames = [os.path.basename(d)
                                for d in glob.glob(zone_glob)
                                if os.path.isdir(d)]
    LOGGER.debug('Zone directory basenames: %s', zone_directory_basenames)

    reporting_config = {
        'title': _('Report'),
        'sortable': False,
        'totals': False,
        'out_uri': out_file,
        'elements': [
            {
                'type': 'head',
                'section': 'head',
                'format': 'style',
                'position': 0,
                'input_type': 'File',
                'data_src': os.path.join(REPORT_DATA, 'table_style.css')
            },
            {
                'type': 'text',
                'section': 'body',
                'input_type': 'text',
                'position': 0,
                'text': (
                    '<h1>{distribution} {title}</h1>'
                    '{help_text}'
                    '<ul>{formatted_hzone_list}</ul>'
                ).format(
                    distribution=distribution.upper(),
                    title=_('Per-Zone Analyses'),
                    help_text=_(
                        'Select a zone to navigate to its impact summary.'
                    ),
                    formatted_hzone_list='\n'.join([
                        '<li><a href="{href}">{text}</li>'.format(
                            href=os.path.join(d, d + '_report.html'),
                            text=d)
                        for d in zone_directory_basenames])
                )
            }
        ]
    }
    reporting.generate_report(reporting_config)

def build_report(municipalities, biodiversity_impact, selected_parcels,
    project_footprint, total_impacts,
    impact_type, output_workspace, impact_sites, pop_col, report_name,
    natural_parcels, impacts_error=False, suggested_parcels=[],
    custom_es_servicesheds=None, dev_dir='_dev', service_mitrat={'carbon': 1.0,
        'nutrient': 1.0, 'sediment': 1.0}, per_offset_data=None,
    prop_offset=1.0, distribution='opal', include_aoi_column=True,
    include_subzone_column=True, hzone_name=''):

    # sort the suggested offset parcels
    suggested_parcels = sorted(suggested_parcels)

    impact_columns = [
        {
            'name': _('Impact footprint shapefile'),
            'total': False,
            'attr':{'class':'round2'}},
    ]

    dev_dir = os.path.join(output_workspace, dev_dir)

    if not os.path.exists(dev_dir):
        os.makedirs(dev_dir)

    # Determine which municipalities the selected polygons overlap with.
    temp_municipalities = os.path.join(output_workspace,
        'servicesheds_with_offset_parcels.shp')
#    per_offset_data = analysis.percent_overlap(selected_parcels,
#        municipalities, temp_municipalities, pop_col)

#    LOGGER.debug('Per-offset data: \n%s',
#        json.dumps(per_offset_data, indent=4, sort_keys=True))
#    json.dump(per_offset_data, open(os.path.join(dev_dir,
#        'per_offset_data.json'), 'w'), indent=4, sort_keys=True)
#    opal_reporting.write_per_offset_csv(per_offset_data,
#            os.path.join(output_workspace, 'offset_benefits_to_servicesheds.csv'))
#    LOGGER.debug('Total impacts: %s', total_impacts)

    biodiversity_data = []
    for ecosystem_name, bio_impact_dict in biodiversity_impact.iteritems():
        try:
            ecosystem_name = unicode(ecosystem_name, 'utf-8')
        except TypeError:
            # happens when trying to decode a unicode string to unicode.
            pass

        ecosystem_impact_dict = {
            _('Ecosystem type'): ecosystem_name,
            _('Total impacted area (ha)'): round(bio_impact_dict['impacted_area'], 2),
            _('Required offset area (ha)'): round(bio_impact_dict['mitigation_area'],
                2),
            _('Mean mitigation ratio'): round(bio_impact_dict['mitigation_ratio'], 2),
            _('No. patches impacted'): bio_impact_dict['patches_impacted'],

            # initialize for the biodiversity impacts table
            _('Area offset (ha)'): 0.0,
            _('Net benefits (ha)'): bio_impact_dict['mitigation_area'] * -1.0,
        }
        biodiversity_data.append(ecosystem_impact_dict)

    # determine whether to skip biodiversity based on the length of the
    # biodiversity impacts dict.
    skip_biodiv = len(biodiversity_impact) == 0
    EMPTY_REPORT_OBJ = lambda : {
        'type': 'text',
        'section': 'body',
        'position': 0,
        'text': '',
    }.copy()
    if not skip_biodiv:
        impacted_parcels_table = opal_reporting.impacted_parcels_table(
            impact_sites, natural_parcels, os.path.join(output_workspace,
            'impacted_natural_ecosystems.csv'))
    else:
        impacted_parcels_table = EMPTY_REPORT_OBJ()

    # get the amount of service required for global services from the
    # impact and the mitigation ratio.
    adjusted_global_impacts = {}
    for service_name, service_impact in total_impacts.iteritems():
        adjusted_impact = service_impact * service_mitrat[service_name]
        adjusted_global_impacts[service_name] = adjusted_impact

    LOGGER.debug('Total impacts before adjustment: %s', total_impacts)
    LOGGER.debug('adjusted_global_impacts: %s', adjusted_global_impacts)
    offset_parcels_table = opal_reporting.build_parcel_table(per_offset_data,
        adjusted_global_impacts.copy(), os.path.join(output_workspace,
        'offset_sites_filtered_table.csv'), distribution, include_aoi_column,
        include_subzone_column, suggested_parcels)

    report_args = {
        'title': _('Report'),
        'sortable': True,
        'totals': True,
        'out_uri': os.path.join(output_workspace, report_name),
        'elements': [
            {
                'type': 'head',
                'section': 'head',
                'format': 'style',
                'position': 0,
                'input_type': 'File',
                'data_src': os.path.join(REPORT_DATA, 'table_style.css')
            },
            {
                'type': 'head',
                'section': 'head',
                'format': 'script',
                'input_type': 'File',
                'data_src': os.path.join(REPORT_DATA, 'permitting_functions.js'),
            },
            {
                'type': 'head',
                'section': 'head',
                'format': 'script',
                'input_type': 'File',
                'data_src': os.path.join(REPORT_DATA, 'toc.min.js'),
            },
            {
                'type': 'head',
                'section': 'head',
                'format': 'script',
                'input_type': 'File',
                'data_src': os.path.join(REPORT_DATA, 'table_row_mouseover.js'),
            },
            {  # add in the JSON data necessary for interactive tables
                'type': 'head',
                'section': 'head',
                'format': 'json',
                'data_src': json.dumps(recurse_sigfig(per_offset_data, 3)),
                'input_type': 'Text',
                'attributes': {'id': 'muni-data'},
            },
            {
                'type': 'head',
                'section': 'head',
                'format': 'json',
                'input_type': 'Text',
                'data_src': json.dumps(opal_reporting.get_impact_data(
                    temp_municipalities, impact_sites, pop_col,
                    'pop_size', service_mitrat,
                    os.path.join(dev_dir, 'serviceshed_data.json'))),
                'attributes': {'id': 'impact-data'},
            },
            {
                'type': 'text',
                'section': 'body',
                'input_type': 'text',
                'position': 0,
                'text': '<div id="TOC">%s</div><div id="content">',
            },
            {
                'type': 'text',
                'section': 'body',
                'input_type': 'text',
                'position': 0,
                'text': ('<div class="notice">%s</div>' %_('<b>NOTE:</b> Impact'
                    ' and offset values are <b>unitless</b> unless you have'
                    ' calibrated your static map model runs.')),
            },
            {
                'type': 'text',
                'section': 'body',
                'position': 0,
                'text': ''.join([
                    '<h1>{hzone}: {title}</h1>'.format(
                        hzone=hzone_name,
                        title=_('Summary of impacts to ecosystems and '
                                'ecosystem services').encode('utf-8')),
                    '<p>{i18n_impact}: <strong>{imp_type}</strong></p>'.format(
                        i18n_impact=_('Impact type').encode('utf-8'),
                        imp_type=impact_type)
                    ])
            },
            {
                'type': 'text',
                'section': 'body',
                'position': 0,
                'text': '<h2>%s</h2>' % _('Total impacts to ecosystem services'),
            },
            opal_reporting.es_impacts_table(total_impacts, service_mitrat),
            {
                'type': 'text',
                'section': 'body',
                'position': 0,
                'text': '<h2>%s</h2>' % _('Total impacts to terrestrial ecosystems'),
            } if not skip_biodiv else EMPTY_REPORT_OBJ(),
            {
                'type': 'table',
                'section': 'body',
                'sortable': True,
                'checkbox': False,
                'total': True,
                'data_type': 'dictionary',
                'key': _('Ecosystem type'),
                'columns': [
                    {
                        'name': _('Ecosystem type'),
                        'total': False
                    },
                    {
                        'name': _('No. patches impacted'),
                        'total': True
                    },
                    {
                        'name': _('Mean mitigation ratio'),
                        'total': False,
                        'attr': {'class': 'round2'},
                    },
                    {
                        'name': _('Total impacted area (ha)'),
                        'total': True,
                        'attr': {'class': 'round2'},
                    },
                    {
                        'name': _('Required offset area (ha)'),
                        'total': True,
                        'attr': {'class': 'round2'},
                    },
                ],
                'data': biodiversity_data,
            } if not skip_biodiv else EMPTY_REPORT_OBJ(),
            {
                'type': 'text',
                'section': 'body',
                'position': 0,
                'text': '<h2>%s</h2>' % _('Details of impacts to terrestrial ecosystems'),
            } if not skip_biodiv else EMPTY_REPORT_OBJ(),
            impacted_parcels_table,
            {
                'type': 'text',
                'section': 'body',
                #'position': 0,
                'text': '<div><h1>%s</h1>' % _('Offset patches and mitigation potential'),
            },
            {
                'type': 'text',
                'section': 'body',
                #'position': 0,
                'text': (
                    '<h2>{title}</h2>{export_csv}<br/><br/>'
                    '{all_offsets_title}: <b>{offsets_link}</b>'
                    '<br/>').format(
                        title=_('Possible offset patches').encode('utf-8'),
                        export_csv=('<a href="#" '
                                    'class="export">Export CSV</a>'),
                        all_offsets_title=_('GIS vector with all selected '
                                            'offset patches').encode('utf-8'),
                        offsets_link=os.path.relpath(
                            selected_parcels,
                            os.path.join(output_workspace, '..', '..')))
            },
            {
                'type': 'text',
                'section': 'body',
                #'position': 0,
                'text': '<b>%s: %s</b><br/>%s<br/>' % (_('Suggested offset parcels').encode('utf-8'),
                    suggested_parcels,
                    (_('Suggested parcels account for %sx impacts x mitigation ratio') % prop_offset).encode('utf-8'))
            },
            offset_parcels_table,
            {
                'type': 'text',
                'section': 'body',
                #'position': 0,
                'text': '</div><div><h1>%s</h1>' % _('Net benefits to ecosystems and ecosystem services'),
            },
            {
                'type': 'text',
                'section': 'body',
                'text': '<h2>%s</h2>' % _('Benefits to terrestrial ecosystems'),
            } if not skip_biodiv else EMPTY_REPORT_OBJ(),
            {
                'type': 'table',
                'section': 'body',
                'sortable': True,
                'total': False,
                'attributes': {'id': 'bio_table'},
                'data_type': 'dictionary',
                'key': _('Ecosystem type'),
                'columns': [
                    {
                        'name': _('Ecosystem type'),
                        'total': False
                    },
                    {
                        'name': _('Total impacted area (ha)'),
                        'total': True
                    },
                    {
                        'name': _('Required offset area (ha)'),
                        'total': False,
                        'attr': {'class': 'required_offset round2'}
                    },
                    {
                        'name': _('Area offset (ha)'),
                        'total': False,
                        'attr': {'class': 'area_offset round2'}
                    },
                    {
                        'name': _('Net benefits (ha)'),
                        'total': False,
                        'attr': {'class': 'net round2'}
                    },
                ],
                'data': biodiversity_data,
            } if not skip_biodiv else EMPTY_REPORT_OBJ(),
            {
                'type': 'text',
                'section': 'body',
                'text': '<h2>%s</h2>' % _('Ecosystem service benefits'),
            },
            {
                'type': 'text',
                'section': 'body',
                'input_type': 'text',
                'position': 0,
                'text': ('<div id="metadata">'
                         'OPAL {opal_version}<br/>'
                         '{timestamp}<br/>'
                         '</div>').format(
                             opal_version=natcap.opal.__version__,
                             timestamp=time.strftime('%Y-%m-%d %H:%M:%S'))
            },
            {
                'type': 'text',
                'section': 'body',
                'text': '</div></div>',
            },
        ]
    }
    if True in map(lambda x: x in total_impacts.keys(), ['sediment',
        'nutrient']) or custom_es_servicesheds == 'hydrological':
        LOGGER.debug('Including the hydrological table')
        hydro_services = []
        for possible_hydro_service in ['sediment', 'nutrient']:
            if possible_hydro_service in total_impacts.keys():
                hydro_services.append(possible_hydro_service)
        if custom_es_servicesheds == 'hydrological':
            hydro_services.append('custom')
        hydrological_table =  opal_reporting.es_benefits_table(hydro_services)
        report_args['elements'].insert(-2, hydrological_table)

    if 'carbon' in total_impacts or custom_es_servicesheds == 'global':
        LOGGER.debug('Including the global table')
        global_table = opal_reporting.global_benefits_table(
            custom_es_servicesheds == 'global', adjusted_global_impacts)
        report_args['elements'].insert(-2, global_table)

    # If the user's impact sites intersect an avoidance area, then insert a
    # warning into the HTML report.
    if impacts_error == True:
        warning = {
            'type': 'text',
            'section': 'body',
            'input_type': 'text',
            'position': 0,
            'text': ('<div class="warning">%s</div>' %_('WARNING: Impact sites'
                ' intersect an avoidance area')),
        }

        # insert the warning just after the impact site summary label.
        report_args['elements'].insert(5, warning)

    reporting.generate_report(report_args)

def write_vector(in_vector_uri, feature_indices, out_vector_uri,
        include_fields=None, new_fields=None, new_values=None, field_types={}):
    in_vector = ogr.Open(in_vector_uri)
    in_layer = in_vector.GetLayer()
    in_layer_srs = in_layer.GetSpatialRef()
    in_layer_defn = in_layer.GetLayerDefn()

    if os.path.exists(out_vector_uri):
        LOGGER.warning('%s already exists on disk', out_vector_uri)
        preprocessing.rm_shapefile(out_vector_uri)

    LOGGER.debug('Creating a new vector at %s', out_vector_uri)
    LOGGER.debug('Out vector type: %s', type(out_vector_uri))
    out_driver = ogr.GetDriverByName('ESRI Shapefile')
    out_vector = out_driver.CreateDataSource(out_vector_uri)

    # Cast to str here because the ESRI shapefile expects it.
    layer_name = str(os.path.basename(os.path.splitext(out_vector_uri)[0]))
    out_layer = out_vector.CreateLayer(layer_name, srs=in_layer_srs)

    # map the old index to the new index.
    field_indices = {}
    if include_fields is None:
        LOGGER.debug('No fields will be copied')
        include_fields = []

    if include_fields is 'all':
        LOGGER.debug('All fields will be copied')
        field_names = []
        for index in range(in_layer_defn.GetFieldCount()):
            field_defn = in_layer_defn.GetFieldDefn(index)
            field_names.append(field_defn.GetName())
        include_fields = field_names

    LOGGER.debug('Copying fields %s', include_fields)
    for old_field in include_fields:
        old_field_index =  in_layer_defn.GetFieldIndex(old_field)
        old_field_defn = in_layer_defn.GetFieldDefn(old_field_index)

        # create a new field in the output layer based on the old field
        # definition
        out_layer.CreateField(old_field_defn)
        out_layer_defn = out_layer.GetLayerDefn()
        new_field_index = out_layer_defn.GetFieldIndex(old_field)
        field_indices[old_field_index] = new_field_index

    # assume new_fields is a list of fieldnames, and all new fields are floats.
    if new_fields is not None:
        LOGGER.debug('Creating new fields %s', new_fields)
        for fieldname in new_fields:
            try:
                field_type = field_types[fieldname]
            except KeyError:
                field_type = ogr.OFTReal
            field_defn = ogr.FieldDefn(fieldname, field_type)
            out_layer.CreateField(field_defn)

    layer_schema = map(lambda d: d.GetName(), out_layer.schema)

    index_map = {}
    for feature_index in feature_indices:
        old_feature = in_layer.GetFeature(feature_index)
        new_feature = ogr.Feature(out_layer_defn)
        new_feature.SetGeometry(old_feature.GetGeometryRef())

        for old_index, new_index in field_indices.iteritems():
            old_value = old_feature.GetField(old_index)
            new_feature.SetField(new_index, old_value)

            if new_values is not None:
                for fieldname, value in new_values[feature_index].iteritems():
                    if fieldname not in layer_schema:
                        continue

                    new_feature.SetField(fieldname, value)

        out_layer.CreateFeature(new_feature)

        index_map[feature_index] = new_feature.GetFID()

    return index_map


def recurse_sigfig(per_offset_data, num_digits):
    """
    Recurse through a dictionary of offset data.  Any impacts to ecosystem
    services will be rounded to `num_digits` significant figures.

    Parameters:
        per_offset_data(dict): A dictionary of values, mapping string parcel
            IDs to dictionaries of values, including keys with service names.
        num_digits(int): The number of digits to round to.

    Returns:
        A dictionary of the same structure as `per_offset_data`.
    """

    service_names = set([_('Sediment'), _('Nitrogen'), _('Carbon'),
                         _('Custom')])

    def _recurse(iterable, key=None):
        if isinstance(iterable, dict):
            return_iterable = {}
            for key, value in iterable.iteritems():
                return_iterable[key] = _recurse(value, key=key)
            return return_iterable
        elif isinstance(iterable, list):
            return [_recurse(item) for item in iterable]
        else:
            if key in service_names:
                return opal_reporting.sigfig(iterable, num_digits)
            else:
                return iterable

    return _recurse(per_offset_data)



import unittest
import tempfile
import os

from osgeo import gdal
from osgeo import osr
import numpy
from invest_natcap.testing import GISTest
import pygeoprocessing

from adept import static_maps

def raster(numpy_matrix, nodata):
    out_uri = pygeoprocessing.temporary_filename()
    datatype = gdal.GDT_Float32
    n_rows, n_cols = numpy_matrix.shape
    driver = gdal.GetDriverByName('GTiff')
    new_raster = driver.Create(out_uri, n_cols, n_rows, 1, datatype)

    # create some projection information based on the GDAL tutorial at
    # http://www.gdal.org/gdal_tutorial.html
    srs = osr.SpatialReference()
    srs.SetUTM(11, 1)
    srs.SetWellKnownGeogCS('NAD27')
    new_raster.SetProjection(srs.ExportToWkt())
    new_raster.SetGeoTransform([444720, 30, 0, 3751320, 0, -30])

    band = new_raster.GetRasterBand(1)
    band.SetNoDataValue(nodata)
    band.WriteArray(numpy_matrix, 0, 0)

    band = None
    new_raster = None
    return out_uri

def get_matrix(raster_uri):
    dataset = gdal.Open(raster_uri)
    band = dataset.GetRasterBand(1)
    matrix = band.ReadAsArray()

    band = None
    dataset = None
    return matrix

class StaticMapsTest(GISTest):
    def setUp(self):
        self._temp_files = []

    def tearDown(self):
        for file_uri in self._temp_files:
            try:
                os.remove(file_uri)
            except OSError:
                # This happens if the file doesn't exist.  This is usually okay,
                # since it may have been deleted by another method.
                pass

    def temporary_filename(self):
        """Create a new temporary file on disk in the system's temp location.
        This test class tracks the temp files created, removing them at the end
        of the test run."""
        file_handle, path = tempfile.mkstemp()
        os.close(file_handle)
        self._temp_files.append(path)
        return path

    def test_convert_lulc(self):
        lulc_matrix = numpy.matrix([
            [5, 5, 5],
            [4, 3, 1],
            [6, 4, 1]])
        lulc_uri = raster(lulc_matrix, 5)

        out_uri = self.temporary_filename()
        static_maps.convert_lulc(lulc_uri, 7, out_uri)

        expected_matrix = numpy.matrix([
            [5, 5, 5],
            [7, 7, 7],
            [7, 7, 7]])
        self.assertMatrixes(get_matrix(out_uri), expected_matrix)

    @unittest.skip('sediment takes forever')
    def test_execute_sediment_smoke(self):
        lulc_matrix = numpy.matrix([
            [5, 5, 5],
            [4, 3, 1],
            [6, 4, 1]])
        lulc_uri = raster(lulc_matrix, 5)

        dem_matrix = numpy.matrix([
            [5, 5, 5],
            [5, 4, 5],
            [4, 3, 2]])
        dem_uri = raster(dem_matrix, -1)

        workspace_folder = pygeoprocessing.temporary_folder()
        static_maps.execute_model('sediment', lulc_uri, workspace_folder)

    def test_execute_carbon_smoke(self):
        lulc_matrix = numpy.matrix([
            [5, 5, 5],
            [4, 3, 1],
            [6, 4, 1]])
        lulc_uri = raster(lulc_matrix, 5)

        workspace_folder = pygeoprocessing.temporary_folder()
        static_maps.execute_model('carbon', lulc_uri, workspace_folder)



import unittest
import os
import shutil

from shapely.geometry import Polygon

from natcap.opal.tests import vector, COLOMBIA_SRS
from natcap.opal.tests import test_smoke
from natcap.opal import offsets

class OffsetTest(unittest.TestCase):
    def test_select_set_multifactor_bio(self):
        parcels_dict = {
            1: {'area': 400, 'carbon': 3500, 'sediment': 290, 'ecosystem': 'a'},
            2: {'area': 123, 'carbon': 382, 'sediment': 1348, 'ecosystem': 'a'},
            3: {'area': 8392, 'carbon': 1910, 'sediment': 18234, 'ecosystem': 'b'},
            4: {'area': 149, 'carbon': 192, 'sediment': 1019, 'ecosystem': 'b'},
        }

        bio_requirements = {
            'a': {
                'mitigation_area': 200,
            },
            'b': {
                'mitigation_area': 100,
            }
        }

        expected_results = [1, 3]  # largest parcels selected first
        selected_parcels = offsets.select_set_multifactor(parcels_dict,
            bio_requirements)
        self.assertEqual(selected_parcels, expected_results)

    def test_select_set_multifactor_bio_no_ecosystem(self):
        parcels_dict = {
            1: {'area': 400, 'carbon': 3500, 'sediment': 290, 'ecosystem': 'a'},
            2: {'area': 123, 'carbon': 382, 'sediment': 1348, 'ecosystem': 'a'},
            3: {'area': 8392, 'carbon': 1910, 'sediment': 18234, 'ecosystem': 'b'},
            4: {'area': 149, 'carbon': 192, 'sediment': 1019, 'ecosystem': 'b'},
        }

        bio_requirements = {
            'a': {
                'mitigation_area': 200,
            },
            'b': {
                'mitigation_area': 100,
            },
            'c': {  # no available parcels for this ecosystem.
                'mitigation_area': 50,
            }
        }

        expected_results = [1, 3]  # largest parcels selected first
        selected_parcels = offsets.select_set_multifactor(parcels_dict,
            bio_requirements)
        self.assertEqual(selected_parcels, expected_results)

    def test_select_set_multifactor_bio_mega(self):
        parcels_dict = {
            1: {'area': 400, 'carbon': 3500, 'sediment': 290, 'ecosystem': 'a'},
            2: {'area': 123, 'carbon': 382, 'sediment': 1348, 'ecosystem': 'a'},
            3: {'area': 8392, 'carbon': 1910, 'sediment': 18234, 'ecosystem': 'b'},
            4: {'area': 149, 'carbon': 192, 'sediment': 1019, 'ecosystem': 'b'},
        }

        bio_requirements = {
            'a': {
                'mitigation_area': 100,
            },
            'b': {
                'mitigation_area': 100,
            }
        }
        prop_offset = 2.0

        expected_results = [1, 3]
        selected_parcels = offsets.select_set_multifactor(parcels_dict,
            bio_requirements, proportion_offset=prop_offset)
        self.assertEqual(selected_parcels, expected_results)

    def test_select_set_multifactor_bio_hydro(self):
        parcels_dict = {
            1: {'area': 400, 'carbon': 3500, 'sediment': 290, 'ecosystem': 'a'},
            2: {'area': 123, 'carbon': 382, 'sediment': 1348, 'ecosystem': 'a'},
            3: {'area': 8392, 'carbon': 1910, 'sediment': 18234, 'ecosystem': 'b'},
            4: {'area': 149, 'carbon': 192, 'sediment': 1019, 'ecosystem': 'b'},
        }

        bio_requirements = {
            'a': {
                'mitigation_area': 200,
            },
            'b': {
                'mitigation_area': 100,
            }
        }

        es_requirements = {
            1: {
                'sediment': 2500,
                'parcels': [1, 2, 3],
            }
        }
        expected_results = [1, 2, 4]
        selected_parcels = offsets.select_set_multifactor(parcels_dict,
            bio_requirements, es_requirements)

    def test_select_set_multifactor_hydro(self):
        parcels_dict = {
            1: {'area': 400, 'carbon': 3500, 'sediment': 290, 'ecosystem': 'a'},
            2: {'area': 123, 'carbon': 382, 'sediment': 1348, 'ecosystem': 'a'},
            3: {'area': 8392, 'carbon': 1910, 'sediment': 18234, 'ecosystem': 'b'},
            4: {'area': 149, 'carbon': 192, 'sediment': 1019, 'ecosystem': 'b'},
        }

        es_requirements = {
            1: {
                'sediment': 4000,
                'parcels': [1, 2, 3],
            }
        }

        expected_parcels = [3]
        selected_parcels = offsets.select_set_multifactor(parcels_dict,
            es_hydro_req=es_requirements)
        self.assertEqual(selected_parcels, expected_parcels)

    def test_select_set_multifactor_hydro_mega(self):
        parcels_dict = {
            1: {'area': 400, 'carbon': 3500, 'sediment': 290, 'ecosystem': 'a'},
            2: {'area': 123, 'carbon': 382, 'sediment': 1348, 'ecosystem': 'a'},
            3: {'area': 8392, 'carbon': 1910, 'sediment': 18234, 'ecosystem': 'b'},
            4: {'area': 149, 'carbon': 192, 'sediment': 1019, 'ecosystem': 'b'},
        }

        es_requirements = {
            1: {
                'sediment': 4000,
                'parcels': [1, 2, 3],
            }
        }

        prop_offset = 4.75

        expected_parcels = [2, 3]
        selected_parcels = offsets.select_set_multifactor(parcels_dict,
            es_hydro_req=es_requirements, proportion_offset=prop_offset)
        self.assertEqual(selected_parcels, expected_parcels)

    def test_select_set_multifactor_global(self):
        parcels_dict = {
            1: {'area': 400, 'carbon': 3500, 'sediment': 290, 'ecosystem': 'a'},
            2: {'area': 123, 'carbon': 382, 'sediment': 1348, 'ecosystem': 'a'},
            3: {'area': 8392, 'carbon': 1910, 'sediment': 18234, 'ecosystem': 'b'},
            4: {'area': 149, 'carbon': 192, 'sediment': 1019, 'ecosystem': 'b'},
        }

        global_reqs = {
            'carbon': 4000,
            'sediment': 500,
        }

        expected_parcels = [1, 3]
        selected_parcels = offsets.select_set_multifactor(parcels_dict,
            es_global_req=global_reqs)
        self.assertEqual(expected_parcels, selected_parcels)

    def test_select_set_multifactor_bio_hydro_global(self):
        parcels_dict = {
            1: {'area': 400, 'carbon': 3500, 'sediment': 290, 'ecosystem': 'a'},
            2: {'area': 123, 'carbon': 382, 'sediment': 1348, 'ecosystem': 'a'},
            3: {'area': 8392, 'carbon': 1910, 'sediment': 18234, 'ecosystem': 'b'},
            4: {'area': 149, 'carbon': 192, 'sediment': 1019, 'ecosystem': 'b'},
        }

        bio_requirements = {
            'a': {
                'mitigation_area': 200,
            },
            'b': {
                'mitigation_area': 100,
            }
        }

        es_requirements = {
            1: {
                'sediment': 2500,
                'parcels': [1, 2, 3],
            }
        }

        global_reqs = {
            'carbon': 4000,
            'sediment': 500,
        }

        expected_parcels = [1, 2, 3, 4]
        selected_parcels = offsets.select_set_multifactor(parcels_dict,
            bio_requirements, es_requirements, global_reqs)
        self.assertEqual(expected_parcels, selected_parcels)

    def test_select_set(self):
        parcels_dict = {
            1: {'area': 400, 'carbon': 3500, 'sediment': 290},
            2: {'area': 123, 'carbon': 382, 'sediment': 1348},
            3: {'area': 8392, 'carbon': 1910, 'sediment': 18234},
            4: {'area': 149, 'carbon': 192, 'sediment': 1019},
        }
        requirements = [
            ('area', 350),
            ('sediment', 2000),
            ('carbon', 100),
        ]

        selected_parcels = offsets.select_set(parcels_dict, requirements)

        # Parcel 3 meets all three requirements.
        self.assertEqual(selected_parcels, [3])

    def test_select_set_2(self):
        parcels = {
            1: {'area': 101, 'carbon': 201, 'sediment': 301},
            2: {'area': 102, 'carbon': 202, 'sediment': 302},
            3: {'area': 103, 'carbon': 203, 'sediment': 303},
            4: {'area': 104, 'carbon': 204, 'sediment': 304},
        }
        requirements = [
            ('area', 100),
            ('carbon', 400),
            ('sediment', 900),
        ]

        self.assertEqual(offsets.select_set(parcels, requirements), [2, 3, 4])

    def test_locate_biodiversity_offsets(self):
        eco_a = test_smoke.square((20, 20), 10)
        eco_b = test_smoke.square((60, 20), 10)
        fields = {
            'ecosystem': str,
            'LCI': float,
        }
        field_values = [
            {'ecosystem': 'eco_a', 'LCI': 0.3},
            {'ecosystem': 'eco_b', 'LCI': 0.2},
        ]
        natural_ecosystems = vector([eco_a, eco_b], COLOMBIA_SRS, fields,
            field_values, format='ESRI Shapefile')
        known_biodiversity_impacts = {
            'eco_a': {
                'min_impacted_parcel_area': 50,
                'min_lci': 0.1,
            },
        }


        expected_parcels = {
            0: {
                'ecosystem': 'eco_a',
                'area': 100.0,
                'lci': 0.3,
            },
        }
        self.assertEqual(offsets.locate_biodiversity_offsets(natural_ecosystems,
            known_biodiversity_impacts), expected_parcels)

    def test_locate_biodiversity_offsets_all(self):
        eco_a = test_smoke.square((20, 20), 10)
        eco_b = test_smoke.square((60, 20), 10)
        fields = {
            'ecosystem': str,
            'LCI': float,
        }
        field_values = [
            {'ecosystem': 'eco_a', 'LCI': 0.3},
            {'ecosystem': 'eco_b', 'LCI': 0.2},
        ]
        natural_ecosystems = vector([eco_a, eco_b], COLOMBIA_SRS, fields,
            field_values, format='ESRI Shapefile')

        expected_parcels = {
            0: {
                'ecosystem': 'eco_a',
                'area': 100.0,
                'lci': 0.3,
            },
            1: {
                'ecosystem': 'eco_b',
                'area': 100.0,
                'lci': 0.2,
            },
        }
        self.assertEqual(offsets.locate_biodiversity_offsets(natural_ecosystems,
            {}), expected_parcels)

    def test_locate_biodiversity_offsets_bio_and_all(self):
        eco_a = test_smoke.square((20, 20), 10)
        eco_b = test_smoke.square((60, 20), 10)
        fields = {
            'ecosystem': str,
            'LCI': float,
        }
        field_values = [
            {'ecosystem': 'eco_a', 'LCI': 0.3},
            {'ecosystem': 'eco_b', 'LCI': 0.2},
        ]
        natural_ecosystems = vector([eco_a, eco_b], COLOMBIA_SRS, fields,
            field_values, format='ESRI Shapefile')
        known_biodiversity_impacts = {
            'eco_a': {
                'min_impacted_parcel_area': 50,
                'min_lci': 0.1,
            },
        }


        expected_parcels = {
            0: {
                'ecosystem': 'eco_a',
                'area': 100.0,
                'lci': 0.3,
            },
            1: {
                'ecosystem': 'eco_b',
                'area': 100.0,
                'lci': 0.2,
            },
        }
        self.assertEqual(offsets.locate_biodiversity_offsets(natural_ecosystems,
            known_biodiversity_impacts, include_all_ecosystems=True),
            expected_parcels)

    def test_select_offsets(self):
        ecosystems_polygons = [test_smoke.square((x, 20), 10)
                               for x in [20, 60, 100, 140]]

        # Screening should screen out this parcel.
        ecosystems_polygons += [test_smoke.square((160, 20), 10)]
        ecosystems_fields = {
            'ecosystem': str,
            'LCI': float,
            'carbon': float,
            'nutrient': float,
            'sediment': float,
        }
        ecosystems_attributes = [
            {'ecosystem': 'eco_a', 'LCI': 0.1, 'carbon': 822.82,
                'nutrient': 357.12, 'sediment': 395.82},
            {'ecosystem': 'eco_b', 'LCI': 0.1, 'carbon': 396.34,
                'nutrient': 934.77, 'sediment': 121.27},
            {'ecosystem': 'eco_a', 'LCI': 0.1, 'carbon': 831.61,
                'nutrient': 136.92, 'sediment': 379.30},
            {'ecosystem': 'eco_b', 'LCI': 0.1, 'carbon': 402.12,
                'nutrient': 829.71, 'sediment': 571.09},
            {'ecosystem': 'eco_b', 'LCI': 0.1, 'carbon': 123.42,
                'nutrient': 14.34, 'sediment': 592.01},
        ]
        ecosystems_vector = vector(ecosystems_polygons, COLOMBIA_SRS,
            ecosystems_fields, ecosystems_attributes, format='ESRI Shapefile')
        biodiversity_impacts = {
                'eco_a': {
                    'min_impacted_parcel_area': 50,
                    'min_lci': 0.1,
                    'max_threat': None,
                    'min_richness': None,
                    'mitigation_area': 125.0,
                }
            }

        impacts_polygons = map(lambda x: test_smoke.square((x, 50), 10), [40,
            80])
        impact_columns = {
            'carbon': float,
            'sediment': float,
            'nutrient': float,
        }
        impact_field_values = [
            {'carbon': 123.43, 'nutrient': 898.12, 'sediment': 901.23},
            {'carbon': 817.93, 'nutrient': 671.58, 'sediment': 173.61},
        ]
        impact_parcels_vector = vector(impacts_polygons, COLOMBIA_SRS,
            impact_columns, impact_field_values, format='ESRI Shapefile')

        output_workspace = os.path.join(os.getcwd(), 'test_select_offsets')
        output_vector = os.path.join(output_workspace, 'output_vector.shp')
        output_json = os.path.join(output_workspace, 'output_json.json')

        if os.path.exists(output_workspace):
            shutil.rmtree(output_workspace)
        os.makedirs(output_workspace)

        offset_tuple = offsets._select_offsets(ecosystems_vector,
            impact_parcels_vector, biodiversity_impacts, output_vector,
            output_json)

        self.assertEqual(offset_tuple, ([0, 2], {}))


    def test_select_offsets_no_impacts(self):
        ecosystems_polygons = map(lambda x: test_smoke.square((x, 20), 10), [20, 60,
            100, 140])
        ecosystems_fields = {
            'ecosystem': str,
            'LCI': float,
            'carbon': float,
            'nutrient': float,
            'sediment': float,
        }
        ecosystems_attributes = [
            {'ecosystem': 'eco_a', 'LCI': 0.1, 'carbon': 822.82,
                'nutrient': 357.12, 'sediment': 395.82},
            {'ecosystem': 'eco_b', 'LCI': 0.1, 'carbon': 396.34,
                'nutrient': 934.77, 'sediment': 121.27},
            {'ecosystem': 'eco_a', 'LCI': 0.1, 'carbon': 831.61,
                'nutrient': 136.92, 'sediment': 379.30},
            {'ecosystem': 'eco_b', 'LCI': 0.1, 'carbon': 402.12,
                'nutrient': 829.71, 'sediment': 571.09},
        ]
        ecosystems_vector = vector(ecosystems_polygons, COLOMBIA_SRS,
            ecosystems_fields, ecosystems_attributes, format='ESRI Shapefile')
        biodiversity_impacts = {
                'eco_a': {
                    'min_impacted_parcel_area': 0,
                    'min_lci': 0,
                    'max_threat': None,
                    'min_richness': None,
                    'mitigation_area': 0,
                },
                'eco_b': {
                    'min_impacted_parcel_area': 0,
                    'min_lci': 0,
                    'max_threat': None,
                    'min_richness': None,
                    'mitigation_area': 0,
                }
            }

        impacts_polygons = map(lambda x: test_smoke.square((x, 50), 10), [40,
            80])
        impact_columns = {
            'carbon': float,
            'sediment': float,
            'nutrient': float,
        }
        impact_field_values = [
            {'carbon': 0.0, 'nutrient': 0.0, 'sediment': 0.0},
            {'carbon': 0.0, 'nutrient': 0.0, 'sediment': 0.0},
        ]
        impact_parcels_vector = vector(impacts_polygons, COLOMBIA_SRS,
            impact_columns, impact_field_values, format='ESRI Shapefile')

        output_workspace = os.path.join(os.getcwd(), 'test_select_offsets')
        output_vector = os.path.join(output_workspace, 'output_vector.shp')
        output_json = os.path.join(output_workspace, 'output_json.json')

        if os.path.exists(output_workspace):
            shutil.rmtree(output_workspace)
        os.makedirs(output_workspace)

        offset_tuple = offsets._select_offsets(ecosystems_vector,
            impact_parcels_vector, biodiversity_impacts, output_vector,
            output_json)

        self.assertEqual(offset_tuple, ([0, 1, 2, 3], {}))

    def test_select_offsets_scheme_es_and_bio(self):
        ecosystems_polygons = map(lambda x: test_smoke.square((x, 20), 10), [20, 60,
            100, 140])
        ecosystems_fields = {
            'ecosystem': str,
            'LCI': float,
            'carbon': float,
            'nutrient': float,
            'sediment': float,
        }
        ecosystems_attributes = [
            {'ecosystem': 'eco_a', 'LCI': 0.1, 'carbon': 822.82,
                'nutrient': 357.12, 'sediment': 395.82},
            {'ecosystem': 'eco_b', 'LCI': 0.1, 'carbon': 396.34,
                'nutrient': 934.77, 'sediment': 121.27},
            {'ecosystem': 'eco_a', 'LCI': 0.1, 'carbon': 831.61,
                'nutrient': 136.92, 'sediment': 379.30},
            {'ecosystem': 'eco_b', 'LCI': 0.1, 'carbon': 402.12,
                'nutrient': 829.71, 'sediment': 571.09},
        ]
        ecosystems_vector = vector(ecosystems_polygons, COLOMBIA_SRS,
            ecosystems_fields, ecosystems_attributes, format='ESRI Shapefile')
        biodiversity_impacts = {
                'eco_a': {
                    'min_impacted_parcel_area': 0,
                    'min_lci': 0,
                    'max_threat': None,
                    'min_richness': None,
                    'mitigation_area': 0,
                },
                'eco_b': {
                    'min_impacted_parcel_area': 0,
                    'min_lci': 0,
                    'max_threat': None,
                    'min_richness': None,
                    'mitigation_area': 0,
                }
            }

        impacts_polygons = map(lambda x: test_smoke.square((x, 50), 10), [40,
            80])
        impact_columns = {
            'carbon': float,
            'sediment': float,
            'nutrient': float,
        }
        impact_field_values = [
            {'carbon': 0.0, 'nutrient': 0.0, 'sediment': 0.0},
            {'carbon': 0.0, 'nutrient': 0.0, 'sediment': 0.0},
        ]
        impact_parcels_vector = vector(impacts_polygons, COLOMBIA_SRS,
            impact_columns, impact_field_values, format='ESRI Shapefile')

        output_workspace = os.path.join(os.getcwd(), 'test_select_offsets')
        output_vector = os.path.join(output_workspace, 'output_vector.shp')
        output_json = os.path.join(output_workspace, 'output_json.json')

        if os.path.exists(output_workspace):
            shutil.rmtree(output_workspace)
        os.makedirs(output_workspace)

        offset_tuple = offsets._select_offsets(ecosystems_vector,
            impact_parcels_vector, biodiversity_impacts, output_vector,
            output_json)

        self.assertEqual(offset_tuple, ([0, 1, 2, 3], {}))

    def translate_percent_overlap_to_sshed_data(self):
        sample_dict = {
            0: {
                "Aoi": 0.0,
                "Area": 24769554.721130043,
                "Carbon": 660488.5930480957,
                "City": 0.0,
                "Distance": 105299.1312844463,
                "Ecosystem": "Bosques naturales",
                "Hydrozone": 0.0,
                "Lci": 0.030022655251212,
                "Nitrogen": 15000.409298000042,
                "Sediment": 52037.28566128289,
                "municipalities": {
                    "CASABE_05893": 1.0,
                    "CONCORDIA_47205": 1.0,
                    "SAN LUIS_47798": 1.0
                }
            },
            4: {
                "Aoi": 0.0,
                "Area": 39053053.690533355,
                "Carbon": 1046098.0106658936,
                "City": 0.0,
                "Distance": 111860.43849765629,
                "Ecosystem": "Bosques naturales",
                "Hydrozone": 0.0,
                "Lci": 0.017458507801602,
                "Nitrogen": 22452.06991350127,
                "Sediment": 60673.52671940386,
                "municipalities": {
                    "CASABE_05893": 0.998,
                    "CONCORDIA_47205": 1.0,
                    "SAN PABLO_13670": 0.001,
                }
            },
        }

        expected_parcels = {
            0: {
                'carbon': 660488.5930480957,
                'nutrient': 15000.409298000042,
                'sediment': 52037.28566128289,
                'area': 24769554.721130043,
                'ecosystem': "Bosques naturales",
                'overlap': {
                    'CASABE_05893': 1.0,
                    'CONCORDIA_47205': 1.0,
                    'SAN LUIS_47798': 1.0,
                }
            },
            4: {
                "area": 39053053.690533355,
                "carbon": 1046098.0106658936,
                "nutrient": 22452.06991350127,
                "sediment": 60673.52671940386,
                "ecosystem": "Bosques naturales",
                'overlap': {
                    'CASABE_05893': 0.998,
                    'CONCORDIA_47205': 1.0,
                    'SAN PABLO_13670': 0.001,
                }
            },
        }

        returned_parcels = offsets.translate_parcel_data(sample_dict)

        self.assertEqual(expected_parcels, returned_parcels)

    def test_translate_es_impacts(self):
        sample_dict = {
            0: {
                "Aoi": 0.0,
                "Area": 24769554.721130043,
                "Carbon": 660488.5930480957,
                "City": 0.0,
                "Distance": 105299.1312844463,
                "Ecosystem": "Bosques naturales",
                "Hydrozone": 0.0,
                "Lci": 0.030022655251212,
                "Nitrogen": 15000.409298000042,
                "Sediment": 52037.28566128289,
                "municipalities": {
                    "CASABE_05893": 1.0,
                    "CONCORDIA_47205": 1.0,
                    "SAN LUIS_47798": 1.0
                }
            },
            4: {
                "Aoi": 0.0,
                "Area": 39053053.690533355,
                "Carbon": 1046098.0106658936,
                "City": 0.0,
                "Distance": 111860.43849765629,
                "Ecosystem": "Bosques naturales",
                "Hydrozone": 0.0,
                "Lci": 0.017458507801602,
                "Nitrogen": 22452.06991350127,
                "Sediment": 60673.52671940386,
                "municipalities": {
                    "CASABE_05893": 0.998,
                    "CONCORDIA_47205": 1.0,
                    "SAN PABLO_13670": 0.001,
                }
            },
        }

        expected_parcels = {
            'CASABE_05893': {
                'carbon': 1704494.4076926573,
                'sediment': 112589.46532724795,
                'nutrient': 37407.575071674306,
            },
            'CONCORDIA_47205': {
                'carbon': 1706586.6037139893,
                'sediment': 112710.81238068675,
                'nutrient': 37452.47921150131,
            },
            'SAN PABLO_13670': {
                'carbon': 1046.0980106658935,
                'sediment': 60.67352671940386,
                'nutrient': 22.45206991350127,
            },
            'SAN LUIS_47798': {
                'carbon': 660488.5930480957,
                'sediment': 52037.28566128289,
                'nutrient': 15000.409298000042,
            }
        }
        returned_parcels = offsets.translate_es_impacts(sample_dict)
        self.assertEqual(expected_parcels, returned_parcels)

    def test_group_offsets_by_sshed(self):
        sample_dict = {
            0: {
                "Aoi": 0.0,
                "Area": 24769554.721130043,
                "Carbon": 660488.5930480957,
                "City": 0.0,
                "Distance": 105299.1312844463,
                "Ecosystem": "Bosques naturales",
                "Hydrozone": 0.0,
                "Lci": 0.030022655251212,
                "Nitrogen": 15000.409298000042,
                "Sediment": 52037.28566128289,
                "municipalities": {
                    "CASABE_05893": 1.0,
                    "CONCORDIA_47205": 1.0,
                    "SAN LUIS_47798": 1.0
                }
            },
            4: {
                "Aoi": 0.0,
                "Area": 39053053.690533355,
                "Carbon": 1046098.0106658936,
                "City": 0.0,
                "Distance": 111860.43849765629,
                "Ecosystem": "Bosques naturales",
                "Hydrozone": 0.0,
                "Lci": 0.017458507801602,
                "Nitrogen": 22452.06991350127,
                "Sediment": 60673.52671940386,
                "municipalities": {
                    "CASABE_05893": 0.998,
                    "CONCORDIA_47205": 1.0,
                    "SAN PABLO_13670": 0.001,
                }
            },
        }

        expected_groupings = {
            "CASABE_05893": [0, 4],
            "CONCORDIA_47205": [0, 4],
            "SAN PABLO_13670": [4],
            "SAN LUIS_47798": [0],
        }
        returned_groupings = offsets.group_offset_parcels_by_sshed(sample_dict)
        self.assertEqual(returned_groupings, expected_groupings)




import os
import shutil
import unittest
import tempfile
import hashlib

from osgeo import ogr
from osgeo import osr
from osgeo import gdal
import shapely
import numpy

GDAL_TYPES = {
    numpy.byte: gdal.GDT_Byte,
    numpy.int16: gdal.GDT_Int16,
    numpy.int32: gdal.GDT_Int32,
    numpy.uint16: gdal.GDT_UInt16,
    numpy.uint32: gdal.GDT_UInt32,
    numpy.float32: gdal.GDT_Float32,
    numpy.float64: gdal.GDT_Float64,
}
NUMPY_TYPES = dict((g, n) for (n, g) in GDAL_TYPES.iteritems())

COLOMBIA_SRS = """PROJCS["MAGNA-SIRGAS / Colombia Bogota zone",
    GEOGCS["MAGNA-SIRGAS",
        DATUM["Marco_Geocentrico_Nacional_de_Referencia",
        SPHEROID["GRS 1980",6378137,298.2572221010002,
            AUTHORITY["EPSG","7019"]],
        TOWGS84[0,0,0,0,0,0,0],
        AUTHORITY["EPSG","6686"]],
        PRIMEM["Greenwich",0],
        UNIT["degree",0.0174532925199433],
        AUTHORITY["EPSG","4686"]],
    PROJECTION["Transverse_Mercator"],
    PARAMETER["latitude_of_origin",4.596200416666666],
    PARAMETER["central_meridian",-74.07750791666666],
    PARAMETER["scale_factor",1],
    PARAMETER["false_easting",1000000],
    PARAMETER["false_northing",1000000],
    UNIT["metre",1,
        AUTHORITY["EPSG","9001"]],
    AUTHORITY["EPSG","3116"]]"""

_COLOMBIA_ORIGIN = (444720, 3751320)
def COLOMBIA_GEOTRANSFORM(x_len, y_len, origin=_COLOMBIA_ORIGIN):
    # in the case of Colombia, expect that x_len is positive, y_len is
    # negative.
    #assert x_len > 0
    #assert y_len < 0
    return [origin[0], x_len, 0, origin[1], 0, y_len]

def test_suite():
    return unittest.TestLoader().discover(os.path.dirname(__file__))

VECTOR_FIELD_TYPES = {
    str: ogr.OFTString,
    float: ogr.OFTReal,
    int: ogr.OFTInteger,
}

def cleanup(uri):
    """Remove the uri.  If it's a folder, recursively remove its contents."""
    if os.path.isdir(uri):
        shutil.rmtree(uri)
    else:
        os.remove(uri)

def get_md5sum(uri):
    """Get the MD5 hash for a single file.  The file is read in a
        memory-efficient fashion.

        uri - a string uri to the file to be tested.

        Returns a string hash for the file."""

    block_size = 2**20
    file_handler = open(uri, 'rb')
    md5 = hashlib.md5()
    while True:
        data = file_handler.read(block_size)
        if not data:
            break
        md5.update(data)
    return md5.hexdigest()

def record_workspace_md5sums(workspace_uri, logfile_uri,
        workspace_varname='self.workspace_uri'):
    """Recurse through the workspace_uri and for every file in the workspace,
    record the filepath and md5sum to the logfile.

    Args:
        workspace_uri (string): A URI to the workspace to analyze

        logfile_uri (string): A URI to the logfile to which md5sums and paths
        will be recorded.

        workspace_varname (string): The string variable name that should be
        used when setting the ws_uri variable initially.

    Returns:
        Nothing.
    """

    logfile = open(logfile_uri, 'w')
    _write = lambda x: logfile.write(x + '\n')
    _write('ws_uri = lambda x: os.path.join(%s, x)' %
        workspace_varname)

    ignore_exts = ['.shx']
    for dirpath, dirnames, filenames in os.walk(workspace_uri):
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            if os.path.splitext(filepath)[-1] in ignore_exts:
                continue

            md5sum = get_md5sum(filepath)
            relative_filepath = filepath.replace(workspace_uri + os.sep, '')
            _write("self.assertMD5(ws_uri('%s'), '%s')" %
                (relative_filepath, md5sum))

def vector(geometries, srs_wkt, fields=None, features=None, format='GeoJSON',
        filename=None):
    """Create a temp OGR vector on disk.

        geometries - a list of Shapely objects.
        srs_wkt - the WKT spatial reference string.
        fields - a python dictionary mapping string fieldname to a python
            primitive type.  Example: {'ws_id': int}
        features - a python dictionary mapping fieldname to field value.  The
            field value's type must match the type defined in the fields input.
            It is an error if it doesn't.
        format - a python string indicating the OGR format to write.  GeoJSON
            is pretty good for most things, but doesn't handle multipolygons
            very well.
        filename=None - None or a python string where the file should be saved.
            If None, the vector will be saved to a temporary folder.

    Returns a string URI to the location of the vector on disk."""

    if fields is None:
        fields = {}

    if features is None:
        features = [{} for i in range(len(geometries))]

    assert len(geometries) == len(features)

    if format == 'GeoJSON':
        ext = 'geojson'
    else:
        # assume ESRI Shapefile
        ext = 'shp'

    if filename is None:
        temp_dir = tempfile.mkdtemp()
        vector_uri = os.path.join(temp_dir, 'shapefile.%s' % ext)
    else:
        vector_uri = filename

    out_driver = ogr.GetDriverByName(format)
    out_vector = out_driver.CreateDataSource(vector_uri)

    layer_name = str(os.path.basename(os.path.splitext(vector_uri)[0]))
    srs = osr.SpatialReference()
    srs.ImportFromWkt(srs_wkt)
    out_layer = out_vector.CreateLayer(layer_name, srs=srs)

    for field_name, field_type in fields.iteritems():
        field_defn = ogr.FieldDefn(field_name, VECTOR_FIELD_TYPES[field_type])
        out_layer.CreateField(field_defn)
    layer_defn = out_layer.GetLayerDefn()

    for shapely_feature, fields in zip(geometries, features):
        new_feature = ogr.Feature(layer_defn)
        new_geometry = ogr.CreateGeometryFromWkb(shapely_feature.wkb)
        new_feature.SetGeometry(new_geometry)

        for field_name, field_value in fields.iteritems():
            #assert type(field_value) == fields[field_name]
            new_feature.SetField(field_name, field_value)
        out_layer.CreateFeature(new_feature)

    out_layer = None
    out_vector = None

    return vector_uri

def raster(pixel_matrix, srs_wkt, geotransform, nodata, filename=None):
    """Create a temp GDAL raster on disk.

        pixel_matrix - a numpy matrix representing pixel values.  The datatype
            of the resulting raster will be determined by the datatype of this
            matrix.
        srs_wkt - the WKT spatial reference string.
        geotransform - a 6-element python list of the geotransform for this
            raster.
        nodata - the nodat value for this raster.
        filename=None - the URI to which this raster should be written. If
            none, a random one will be generated.

    Returns a URI to the resulting GDAL raster on disk."""

    if filename is None:
        _, out_uri = tempfile.mkstemp()
        out_uri += '.tif'
    else:
        out_uri = filename

    datatype = GDAL_TYPES[pixel_matrix.dtype.type]
    n_rows, n_cols = pixel_matrix.shape
    driver = gdal.GetDriverByName('GTiff')
    new_raster = driver.Create(out_uri, n_cols, n_rows, 1, datatype)

    # create some projection information based on the GDAL tutorial at
    # http://www.gdal.org/gdal_tutorial.html
    srs = osr.SpatialReference()
    srs.ImportFromWkt(srs_wkt)
    new_raster.SetProjection(srs.ExportToWkt())
    new_raster.SetGeoTransform(geotransform)

    band = new_raster.GetRasterBand(1)
    band.SetNoDataValue(nodata)
    band.WriteArray(pixel_matrix, 0, 0)

    band = None
    new_raster = None
    return out_uri






import unittest

from natcap.opal import reporting

class SigFigTest(unittest.TestCase):
    """
    Tests for significant figure calculations on various numbers.
    """
    def test_sigfig_integer(self):
        output_num = reporting.sigfig(12345, 3)
        self.assertEqual(output_num, 12300)

    def test_sigfig_ceiling(self):
        output_num = reporting.sigfig(12369, 3)
        self.assertEqual(output_num, 12400)

    def test_sigfig_float(self):
        output_num = reporting.sigfig(1.1234, 3)
        self.assertEqual(output_num, 1.12)

    def test_sigfig_largefloat(self):
        output_num = reporting.sigfig(12345.5678, 4)
        self.assertEqual(output_num, 12350.0)

    def test_sigfig_large_negative(self):
        output_num = reporting.sigfig(-12345.34, 3)
        self.assertEqual(output_num, -12300.0)




# -*- coding: utf-8 -*-

import os, sys
import locale
import gettext
import logging

LOGGER = logging.getLogger('natcap.opal.i18n')

# Change this variable to your app name!
#  The translation files will be under
#  @LOCALE_DIR@/@LANGUAGE@/LC_MESSAGES/@APP_NAME@.mo
APP_NAME = "adept"

# This is ok for maemo. Not sure in a regular desktop:
# assume we're executing from the adept source root.
LOCALE_DIR = os.path.join(os.path.dirname(__file__), 'locale') # .mo files will then be located
                                           #in APP_Dir/i18n/LANGUAGECODE/LC_MESSAGES/

# Now we need to choose the language. We will provide a list, and gettext
# will use the first translation available in the list
#
#  In maemo it is in the LANG environment variable
#  (on desktop is usually LANGUAGES)
DEFAULT_LANGUAGES = os.environ.get('LANG', '').split(':')
DEFAULT_LANGUAGES += ['en_US']

lc, encoding = locale.getdefaultlocale()
if lc:
    languages = [lc]

# Concat all languages (env + default locale),
#  and here we have the languages and location of the translations
languages += DEFAULT_LANGUAGES
mo_location = LOCALE_DIR

# Lets tell those details to gettext
#  (nothing to change here for you)
gettext.install(APP_NAME, localedir=None, unicode=1)

gettext.find(APP_NAME, mo_location)

gettext.textdomain(APP_NAME)

gettext.bind_textdomain_codeset(APP_NAME, "UTF-8")

LOGGER.debug('Available languages: %s', languages)
#language = gettext.translation(APP_NAME, mo_location,
#    languages=languages, fallback=True)

# KEEP FALLBACK=TRUE!
# This module is imported when running setup.py, but the language translation
# files have not been loaded to the correct place.  If Fallback=False, an
# exception will be raised at that time.
get_trans = lambda code: gettext.translation(APP_NAME, mo_location,
        languages=[code], fallback=True)

class Language(object):
    def __init__(self):
        self.current_lang = 'en'

        self.available_langs = {
            'en': get_trans('en'),
            'es': get_trans('es'),
        }

    def __call__(self):
        print 'current language is %s' % self.current_lang
        current_language = self.available_langs[self.current_lang]
        current_language.install()
        return current_language

    def _lang(self):
        return self.available_langs[self.current_lang]

    def set(self, lang_code):
        self.current_lang = lang_code
        print 'setting current lang to %s' % lang_code
        self.__call__()  # install the new language.

    def ugettext(self, message):
        self._lang().install()
        return self._lang().ugettext(message)

    def info(self):
        return self._lang().info()

language = Language()

def current_lang():
    """Return the string language code of the currently-active language."""
    return language.info()['language']

def os_default_lang():
    """Return the OS default language."""
    return locale.getdefaultlocale()[0].split('_')[0].lower()
